from pancake.utils import rev_comp, get_new_alignment, new_ops_from_ops, ali_refinement
from pancake.editops import editops_from_seqs


class SharedFeature:

	def __init__(self, reference, entries=None):
		self.seq = reference #either a sequence or an automat
		self.entries = entries if entries else set() #set of FeatureInstances

	def __str__(self):
		out = self.seq[:50] + '...' + self.seq[-50:] + ' ' + str(len(self.seq)) + 'bp\n' if len(self.seq) >200 else self.seq + ' ' + str(len(self.seq)) + 'bp\n'
		for fi in self.entries:
			out += '\t' + str(fi) + '\t' + str(fi.seq) +'\n'
		return out

	def add(self, entry):
		self.entries.add(entry)

	def remove(self, entry):
		self.entries.remove(entry)

	def __len__(self):
		return len(self.entries)

	def __iter__(self):
		for entry in self.entries:
			yield entry

	def turn(self):
		self.seq = rev_comp(self.seq)
		for fi in self.entries:
			fi.seq.turn()

	def alignments_to_entry(self, ref_fi):
		'''return dictionary: alignments between each FI and the given FI'''
		'''called for s_start_SF while SF merging'''
		'''given ref_fi is always original sequence (forward)''' 
		if ref_fi in self.entries:
			ref_seq = ref_fi.seq.seq(self.seq)
			result_dic = {}
			for fi in self.entries:
				if fi != ref_fi:
					ref_ops = list(ref_fi.seq.ops)
					new_ops = new_ops_from_ops(ref_ops, list(fi.seq.ops), self.seq) #ops for alignment fi.seq rebased on ref_fi.seq
					new_ops = ali_refinement(ref_seq,new_ops)
					result_dic[fi] = new_ops
		return result_dic


	def treat_empty_seq(self):
		entries = list(self.entries)
		ref_entry = entries[0]
		entries = entries[1:]
		ref_seq = ref_entry.seq.seq('')
		ref_entry.seq.ops = [len(ref_seq)]
		for entry in entries:
			entry.seq = editops_from_seqs(ref_seq, entry.seq.seq(entry.sh_feat.seq), reverse=entry.seq.reverse)
		self.seq = ref_seq

	def singleton_sf(self, chroms):
		'''returns True if all chromosomes given in set chroms are NOT covered, i.e. SF is a valid part of a singleton region'''
		for entry in self:
			if entry.chrom in chroms:
				return False
		return True


if __name__ == "__main__":
	pass



from pancake.utils import rev_comp

#from utils import rev_comp

class FeatureInstance:

	def __init__(self, chrom, start=None, stop=None, seq =None, sh_feat=None, nextFI=None, prevFI=None):
		self.sh_feat = sh_feat
		self.start = start
		self.stop = stop
		self.nextFI = nextFI
		self.prevFI = prevFI
		self.chrom = chrom
		self.seq= seq #either sequence, or EditOps


	def __str__(self):
		out = '{}\t{}-{}\t'.format(self.chrom, str(self.start), str(self.stop))
		out += '-' if self.sh_feat and self.seq.reverse else '+'
		#out += '\t' + str(fi.seq)
		return out


	def sequence(self):
		#returns original genomic sequence between start and stop
		if self.sh_feat:
			s = self.seq.seq(self.sh_feat.seq) if not self.seq.reverse else rev_comp(self.seq.seq(self.sh_feat.seq))
			return s
		else:
			return self.seq

if __name__ == "__main__":
	pass



import sys
import os
from pancake.sf import SharedFeature

def get_gv(pangenome, chroms, starts, stops, all_flag, region_flag, max_entries, max_nodes, max_edges, output):
	#all_flag: include all genomes
	#region_flag: only show FIs covering given region

	#PREPARATION
	color_dic = {}
	sf_ind, fi_ind = 1,1
	all_chroms = list(pangenome.chromosomes.keys())
	starts_and_stops = {}
	for i in range(len(all_chroms)):
		color_dic[all_chroms[i]] = i * 1/len(all_chroms)*0.5 + 0.05 * 1/len(all_chroms)
	str_dic = dict() #will contain string representation of SFs
	tmp_sfs = dict() # will contain truncated SFs if all_flag unset
	for i in range(len(chroms)): 
		starts_and_stops[chroms[i]] = (starts[i], stops[i])

	#GET SFS and STRING REPRESENTATION 
	for i in range(len(chroms)):
		fi = pangenome.crossing_fi(chroms[i], starts[i])
		if fi.sh_feat and fi.sh_feat not in tmp_sfs:
			if all_flag:
				tmp_sfs[fi.sh_feat] = fi.sh_feat
			else:
				#get dummy SF conatining exclusively FIs appearing in output
				tmp_sfs[fi.sh_feat] = truncate_sf(fi.sh_feat, starts_and_stops, region_flag) 
			sf= tmp_sfs[fi.sh_feat]
			if len(sf) > max_entries:
				str_dic[sf] = 'SF{} [label="{}", shape=rectangle, style=filled, fillcolor="{}"];'.format(sf_ind, long_sf_str(sf), get_sf_color(fi.sh_feat, color_dic))
			else:
				str_dic[sf] = 'SF{} [label="{}", shape=rectangle, style=filled, fillcolor="{}"];'.format(sf_ind, short_sf_str(sf), get_sf_color(fi.sh_feat, color_dic))
			sf_ind+=1 #index for DOT file
		elif not fi.sh_feat:
			str_dic[fi] = 'FI{} [style=filled, label="{}", fillcolor="{} 0.3 0.85"];'.format(fi_ind, gv_str(fi), color_dic[fi.chrom])
			fi_ind+=1 #index for DOT file
		while fi.nextFI and fi.nextFI.start <= stops[i]:
			fi=fi.nextFI
			if fi.sh_feat and fi.sh_feat not in tmp_sfs:
				if all_flag:
					tmp_sfs[fi.sh_feat] = fi.sh_feat
				else:
					tmp_sfs[fi.sh_feat] = truncate_sf(fi.sh_feat, starts_and_stops, region_flag)
				sf= tmp_sfs[fi.sh_feat]
				if len(sf) > max_entries:
					str_dic[sf] = 'SF{} [label="{}", shape=rectangle, style=filled, fillcolor="{}"];'.format(sf_ind, long_sf_str(sf), get_sf_color(fi.sh_feat, color_dic))
				else:
					str_dic[sf] = 'SF{} [label="{}", shape=rectangle, style=filled, fillcolor="{}"];'.format(sf_ind, short_sf_str(sf), get_sf_color(fi.sh_feat, color_dic))
				sf_ind+=1 #index for DOT file
			elif not fi.sh_feat:
				str_dic[fi] = 'FI{} [style=filled, label="{}", fillcolor="{} 0.3 0.85"];'.format(fi_ind, gv_str(fi), color_dic[fi.chrom])
				fi_ind+=1 #index for DOT file

	#check max_nodes
	nnodes = len(str_dic)
	if nnodes>max_nodes: raise ValueError('Number of nodes in resulting graph {} > MAX_NODES {} ...ABORTING'.format(nnodes, max_nodes))

	#GET EDGES
	edge_dict = dict()
	edge_dict['others'] = set()
	for node in str_dic:
		if type(node) == SharedFeature:
			sf=node
			for entry in node:
				if entry.chrom in chroms and entry.chrom not in edge_dict: edge_dict[entry.chrom] = set()
				if entry.nextFI in str_dic:
					#SF to FI
					if entry.chrom in chroms:
						if not (entry.start > starts_and_stops[entry.chrom][1] or entry.stop < starts_and_stops[entry.chrom][0]):
							#colored edge
							edge_dict[entry.chrom].add('\"{}\" -> \"{}\" [penwidth=8 color= "{} 0.3 0.85"];'.format(str_dic[sf].split()[0], str_dic[entry.nextFI].split()[0], color_dic[entry.chrom]))
						else:
							edge_dict['others'].add('\"{}\" -> \"{}\"'.format(str_dic[sf].split()[0], str_dic[entry.nextFI].split()[0]))
					else:
						edge_dict['others'].add('\"{}\" -> \"{}\"'.format(str_dic[sf].split()[0], str_dic[entry.nextFI].split()[0]))
				elif entry.nextFI and entry.nextFI.sh_feat in tmp_sfs:
					#SF to SF
					next_sf = tmp_sfs[entry.nextFI.sh_feat]
					if entry.chrom in chroms:
						if not (entry.start > starts_and_stops[entry.chrom][1] or entry.stop < starts_and_stops[entry.chrom][0]):
							#colored edge
							edge_dict[entry.chrom].add('\"{}\" -> \"{}\" [penwidth=8 color= "{} 0.3 0.85"];'.format(str_dic[sf].split()[0], str_dic[next_sf].split()[0], color_dic[entry.chrom]))
						else:
							edge_dict['others'].add('\"{}\" -> \"{}\"'.format(str_dic[sf].split()[0], str_dic[next_sf].split()[0]))
					else:
						edge_dict['others'].add('\"{}\" -> \"{}\"'.format(str_dic[sf].split()[0], str_dic[next_sf].split()[0]))
		else:
			#FI
			fi = node
			if fi.chrom in chroms and fi.chrom not in edge_dict: edge_dict[fi.chrom] = set()
			if fi.nextFI in str_dic:
				pass
				#should never be the case
			elif fi.nextFI and fi.nextFI.sh_feat in tmp_sfs:
				next_sf = tmp_sfs[fi.nextFI.sh_feat]
				if fi.chrom in chroms:
					if not (fi.start > starts_and_stops[fi.chrom][1] or fi.stop < starts_and_stops[fi.chrom][0]):
						#colored edge
						edge_dict[fi.chrom].add('\"{}\" -> \"{}\" [penwidth=8 color= "{} 0.3 0.85"];'.format(str_dic[fi].split()[0], str_dic[next_sf].split()[0], color_dic[fi.chrom]))
					else:
						edge_dict['others'].add('\"{}\" -> \"{}\"'.format(str_dic[fi].split()[0], str_dic[next_sf].split()[0]))
				else:
					edge_dict['others'].add('\"{}\" -> \"{}\"'.format(str_dic[fi].split()[0], str_dic[next_sf].split()[0]))



	#check max_edges
	nedges = 0
	for key in edge_dict:
		nedges+= len(edge_dict[key])
	if nedges>max_edges: raise ValueError('Number of edges in resulting graph {} > MAX_EDGES {} ...ABORTING'.format(nedges, max_edges))
	
	sys.stderr.write('...resulting graph consists of {} nodes and {} edges\n#\n'.format(nnodes, nedges))
	if not output:
		#print to STDOUT
		print('digraph G {')
		print('graph [fontsize=100 labelloc="t" label="" splines=true overlap=false];\nratio = auto;\n') # rankdir = "TB"
		for key in str_dic:
			print(str_dic[key])
			str_dic[key] = str_dic[key].split()[0]
		for key in edge_dict:
			if key == 'others':
				for edge in edge_dict[key]:
					print(edge + ' [color="0.56 0.15 0.83"];')
			else:
				for edge in edge_dict[key]:
					print(edge)
		print('}')
	else:
		#write to file
		dir, f = os.path.split(output)
		if dir:
			try:
				os.stat(dir)
			except:
				os.mkdir(dir)
		with open(output, 'wt') as dot_file:
			dot_file.write('digraph G {\n')
			dot_file.write('graph [fontsize=100 labelloc="t" label="" splines=true overlap=false];\nratio = auto;\n\n') # rankdir = "TB"
			for key in str_dic:
				dot_file.write(str_dic[key]+'\n')
				str_dic[key] = str_dic[key].split()[0]
			for key in edge_dict:
				if key == 'others':
					for edge in edge_dict[key]:
						#dot_file.write(edge + ' [color="0.56 0.15 0.83"];\n')
						dot_file.write(edge + ' [color="0 0 0.7"];\n')
				else:
					for edge in edge_dict[key]:
						dot_file.write(edge + '\n')
			dot_file.write('}\n')
		sys.stderr.write('...written to DOT file {}\n'.format(output))


def truncate_sf(sf, starts_and_stops, region_flag):
	#all_flag unset 
	new_sf = SharedFeature(sf.seq)
	entry_count, entry_dic = 0, {}
	for entry in sf.entries:
		if entry.chrom in starts_and_stops:
			if not region_flag:
				new_sf.add(entry)
			elif not (entry.start > starts_and_stops[entry.chrom][1] or entry.stop < starts_and_stops[entry.chrom][0]):
				new_sf.add(entry)
	return new_sf


def long_sf_str(sf):
	sf_dict = dict()
	s = sf.seq if len(sf.seq)<=10 else sf.seq[:5] + '...' + sf.seq[-5:]
	s += ' \({}bp\)\\n'.format(len(sf.seq))
	for entry in sf.entries:
		genome = entry.chrom.genome
		if genome not in sf_dict:
			sf_dict[genome] = dict()
		if entry.chrom not in sf_dict[genome]:
			sf_dict[genome][entry.chrom] = 1
		else: sf_dict[genome][entry.chrom] +=1
	for genome in sf_dict:
		s+= 'Genome ' + genome + ':\\n'
		for c in sf_dict[genome]:
			s+= ' ' + str(sf_dict[genome][c]) + 'x ' + str(c) + '\\n'
	return s


def short_sf_str(sf):
	sf_dict = dict()
	s = sf.seq if len(sf.seq)<=10 else sf.seq[:5] + '...' + sf.seq[-5:]
	s += ' \({}bp\)\\n'.format(len(sf.seq))
	for entry in sorted(sf.entries, key=keyfunc):
		s += '{} {}-{} '.format(entry.chrom, entry.start, entry.stop)
		s+= '-' if entry.seq.reverse else '+'
		s+= ' ({}bp)\\n'.format(entry.stop-entry.start+1)
	return s


def gv_str(node, starts_and_stops=None, all_flag=False, max_entries=500):
	s=''
	if type(node) == SharedFeature:
		s += node.seq if len(node.seq)<=10 else node.seq[:5] + '...' + node.seq[-5:]
		s += ' \({}bp\)\\n'.format(len(node.seq))
		entry_count, entry_dic = 0, {}
		for entry in sorted(node.entries, key=keyfunc):
			if entry.chrom not in entry_dic: entry_dic[entry.chrom] = []
			entry_dic[entry.chrom].append(entry)
		for chrom in entry_dic:
			for entry in entry_dic[chrom]:
				s += '{} {}-{} '.format(entry.chrom, entry.start, entry.stop)
				s+= '-' if entry.seq.reverse else '+'
				s+= ' ({}bp)\\n'.format(entry.stop-entry.start+1)
	else:
		s += '{} {}-{} '.format(node.chrom, node.start, node.stop)
		s+= '+ \({}bp\)\\n'.format(node.stop-node.start+1)
	return s


def keyfunc(x):
	return (x.chrom.name, x.start)


def get_sf_color(sf, color_dic):
	color_set=set()
	for f in sf.entries:
		color_set.add(color_dic[f.chrom])
	if len(color_set)>1:
		return "{} 0.6 0.85".format(sum(color_set)/len(color_set) + len(color_set)*0.04)#*min(color_set))
	else:
		return "{} 0.3 0.85".format(list(color_set)[0])




import os
import sys
from Bio import SeqIO,Entrez
import cProfile

_libdir = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../')
if os.path.isfile(os.path.join(_libdir, 'pancake', '__init__.py')):
	sys.path.insert(0, _libdir)

from pancake.pangenome import PanGenome,decompress
from pancake.__init__ import *


def test_sequences():
	pan_file = 'tests/test.pan'
	seq_dict = {}
	create([], ["372102604", '38231477', '372116212'], email=None, pan_file=pan_file)
	pangenome=decompress(pan_file)
	for chrom in pangenome.chromosomes:
		seq_dict[chrom.name] = pangenome.seq(chrom)
	add_alignments(pan_file, ['tests/out.delta'], pan_file, min_len=100, self_alis=False)
	pangenome=decompress(pan_file)
	for chrom in pangenome.chromosomes:
		assert pangenome.seq(chrom) == seq_dict[chrom.name], 'Error in test_sequences after alignment inegration! {}'.format(chrom.name)
	pangenome.compress(pan_file)
	include_genome_info(pan_file, 'tests/genomes.txt')


if __name__ == "__main__":
	pass



import os
import sys

from Bio import SeqIO,Entrez
from pancake.pangenome import PanGenome, decompress
from pancake.aliparser import parse_alis
from pancake.core import get_genome_core,get_core,write_core_fasta_files
from pancake.singletons import get_genome_singletons,get_singletons
from pancake.gv_graphs import get_gv


__author__ = "Corinna Ernst"
__version__ = "1.1"


def create(files, identifiers, email, pan_file=None, alis=None, min_len=25, self_alis=True):

	if not pan_file:
		pan_file = os.getcwd() + '/pan_files/pancake.pan'
		dir = os.path.dirname(pan_file)
		try:
			os.stat(dir)
		except:
			os.mkdir(dir)
	else:
		dir, f = os.path.split(pan_file)
		if dir:
			try:
				os.stat(dir)
			except:
				os.mkdir(dir)
	
	pangenome = PanGenome()
	if files:
		for f in files:
			records = list(SeqIO.parse(open(f, 'r'), "fasta"))
			for record in records:
				pangenome.add_chromosome(str(record.name), seq=str(record.seq))
	if identifiers:
		if email: Entrez.email=email
		for identifier in identifiers:
			handle = Entrez.efetch(db="nucleotide", id=identifier, rettype="fasta", retmode="text")
			record = SeqIO.read(handle, "fasta")
			handle.close()
			pangenome.add_chromosome(str(record.name), seq=str(record.seq).upper())
	sys.stderr.write('#\n...pangenome initialized\n')
	try:
		pangenome.compress(pan_file)
	except:
		sys.stderr.write('Error while pangenome compressing\n')
	if alis:
		for ali_file in alis:
			sys.stderr.write('#\n')
			pangenome = parse_alis(ali_file, pangenome, min_len, self_alis)
			sys.stderr.write('#\n')
			sys.stderr.write('...check for non-similarity SFs\n')
			pangenome.check_non_similarity()
			sys.stderr.write('...downgrade non-aligned Feature Instances\n')
			pangenome.refine_unlabeled()
			sys.stderr.write('...concatenate Shared Features\n')
			pangenome.refine_adjacent_sfs()
			try:
				pangenome.compress(pan_file)
			except:
				sys.stderr.write('Error while pangenome compressing\n')
	statinfo = os.stat(pan_file)
	sys.stderr.write('...PanCake Object written to {} ({} bytes)\n'.format(pan_file, statinfo.st_size))


def add_alignments(pan_file, alis, output, min_len, self_alis=True):

	dir, f = os.path.split(output)
	if dir:
		try:
			os.stat(dir)
		except:
			os.mkdir(dir)

	try:
		pangenome=decompress(pan_file)
	except:
		sys.stderr.write('...unable to initialize pangenome from {}!\n'.format(pan_file))
		return
	sys.stderr.write('...pangenome loaded from {}\n'.format(pan_file))
	for ali_file in alis:
		sys.stderr.write('#\n')
		try:
			pangenome = parse_alis(ali_file, pangenome, min_len, self_alis)
		except:
			sys.stderr.write('Error while parsing {}.\n'.format(ali_file))
		sys.stderr.write('#\n')
		sys.stderr.write('...check for non-similarity SFs\n')
		pangenome.check_non_similarity()
		sys.stderr.write('...downgrade non-aligned Feature Instances\n')
		pangenome.refine_unlabeled()
		sys.stderr.write('...concatenate Shared Features\n')
		pangenome.refine_adjacent_sfs()
		sys.stderr.write('#\n')
		
		try:
			pangenome.compress(output)
		except:
			sys.stderr.write('Error while pangenome compressing\n')
	statinfo = os.stat(output)
	sys.stderr.write('...PanCake Object written to {} ({} bytes)\n'.format(output, statinfo.st_size))


def core(pan_file, ref_chrom, ref_genome, min_len, genomes=[], chroms=[], exclude_genomes=[], exclude_chromosomes=[], non_core_frac=0.1, nc_space=25, folder=None, output=True, bed_file=''):
	try:
		pangenome=decompress(pan_file)
	except:
		sys.stderr.write('...unable to initialize pangenome from {}!\n'.format(pan_file))
		return
	if ref_chrom:
		if not bed_file: bed_file = 'core_{}.bed'.format(ref_chrom)
		open(bed_file, 'w').close()
		print('#\n...parsing ' + ref_chrom)
		regions = get_core(pangenome, ref_chrom, min_len, genomes, chroms, exclude_genomes, exclude_chromosomes, non_core_frac, nc_space, {},bed_file)
	elif ref_genome:
		if not bed_file: bed_file = 'core_{}.bed'.format(ref_genome)
		open(bed_file, 'w').close()
		regions = get_genome_core(pangenome, ref_genome, min_len, genomes, chroms, exclude_genomes, exclude_chromosomes, non_core_frac, nc_space, bed_file)
	if output:
		if not folder: 
			folder = 'core_'+ ref_chrom if ref_chrom else 'core_'+ ref_genome
		write_core_fasta_files(pangenome, regions, ref_chrom, folder)


def singletons(pan_file, ref_chrom, ref_genome, min_len, genomes=[], chroms=[], exclude_genomes=[], exclude_chromosomes=[], outfolder='', output=True, bed_file=''):
	try:
		pangenome=decompress(pan_file)
	except:
		sys.stderr.write('...unable to initialize pangenome from {}!\n'.format(pan_file))
		return
	if ref_chrom:
		if not bed_file: bed_file = 'singletons_{}.bed'.format(ref_chrom)
		open(bed_file, 'w').close()
		get_singletons(pangenome, ref_chrom, min_len, genomes, chroms, exclude_genomes, exclude_chromosomes, outfolder, output, bed_file)
	elif ref_genome:
		if not bed_file: bed_file = 'singletons_{}.bed'.format(ref_genome)
		open(bed_file, 'w').close()
		get_genome_singletons(pangenome, ref_genome, min_len, genomes, chroms, exclude_genomes, exclude_chromosomes, outfolder, output, bed_file)


def stats(pan_file):
	try:
		pangenome=decompress(pan_file)
	except:
		sys.stderr.write('...unable to initialize pangenome from {}!\n'.format(pan_file))
		return
	ul = 0
	ul_dict = {}
	for chrom in pangenome.chromosomes:
		ul+=len(pangenome.unlabeled[chrom])
		ul_dict[chrom] = len(pangenome.unlabeled[chrom])
	l, s = 0, 0
	for f in pangenome.features:
		l+= len(f.entries)
		s+=1
	print('\nPanGenome Object consists of {} un-aligned FIs & {} aligned FIs (organized in {} Shared Features)'.format(ul, l, s))
	print('#\n{} chromosomes representing {} genomes, namely:'.format(len(pangenome.chromosomes), len(pangenome.genomes)))
	#chromosome information
	for genome in pangenome.genomes:
		print('#\nGenome '+str(genome))
		for chrom in pangenome.genomes[genome]:
			ref_name = chrom.name
			names=set()
			for name in pangenome.chrom_names:
				if pangenome.chrom_names[name] == chrom and name != chrom.name: names.add(name)
			if names: 
				print('>'+ref_name + ', ' + ', '.join(list(names)) + ' ('+ str(pangenome.chrom_length(chrom)) +'bp)')
			else: print('>'+ref_name + ' ('+ str(pangenome.chrom_length(chrom)) +'bp)')
			l, l_length,sf_set = set(), 0, set()
			ul_length, ul = 0,0
			fi = pangenome.chromosomes[chrom][1]
			while fi:
				if fi.sh_feat:
					l.add(fi)
					l_length += (fi.stop-fi.start+1)
					sf_set.add(fi.sh_feat)
				else:
					ul_length += (fi.stop-fi.start+1)
					ul +=1
				fi=fi.nextFI
			if ul:
				print('--> {} un-aligned Feature Instances (mean length {})'.format(ul, ul_length/ul))
			else:
				print('--> {} un-aligned Feature Instances (mean length {})'.format(ul, 0))
			if l:
				print('--> {} aligned Feature Instances (mean length {}) in {} Shared Features'.format(len(l), l_length/len(l), len(sf_set)))
			else:
				print('--> {} aligned Feature Instances (mean length {}) in {} Shared Features'.format(0, 0, 0))


def add_chromosomes(panfile, files, ids, email='', outfile='', alis=None, min_len=25, self_alis=True):
	try:
		pangenome = decompress(panfile)
	except:
		raise ValueError('...unable to open {}!'.format(panfile))

	if outfile != panfile:
		dir, f = os.path.split(outfile)
		if dir:
			try:
				os.stat(dir)
			except:
				os.mkdir(dir)
	
	if files:
		for f in files:
			records = list(SeqIO.parse(open(f, 'r'), "fasta"))
			for record in records:
				pangenome.add_chromosome(str(record.name), seq=str(record.seq))
	if ids:
		if email: Entrez.email=email
		for identifier in ids:
			handle = Entrez.efetch(db="nucleotide", id=identifier, rettype="fasta", retmode="text")
			record = SeqIO.read(handle, "fasta")
			handle.close()
			pangenome.add_chromosome(str(record.name), seq=str(record.seq).upper())
	sys.stderr.write('#\n...chromosome(s) added\n')
	if alis:
		for ali_file in alis:
			sys.stderr.write('#\n')
			pangenome = parse_alis(ali_file, pangenome, min_len, self_alis)
			sys.stderr.write('#\n')
			sys.stderr.write('...check for non-similarity SFs\n')
			pangenome.check_non_similarity()
			sys.stderr.write('...downgrade non-aligned Feature Instances\n')
			pangenome.refine_unlabeled()
			sys.stderr.write('...concatenate Shared Features\n')
			pangenome.refine_adjacent_sfs()
			sys.stderr.write('Error while SF concatenation.\n')
			sys.stderr.write('#\n')
	try:
		pangenome.compress(outfile)
	except:
		sys.stderr.write('Error while pangenome compressing\n')
	statinfo = os.stat(outfile)
	sys.stderr.write('...PanCake Object written to {} ({} bytes)\n'.format(outfile, statinfo.st_size))


def delete_name(panfile, names):
	try:
		pangenome = decompress(panfile)
	except:
		raise ValueError('...unable to open {}!'.format(panfile))
	for name in names:
		try:
			chrom = pangenome.chrom_names[name]
			chrom_names=list(filter(lambda x: pangenome.chrom_names[x]==chrom, pangenome.chrom_names))
			if len(chrom_names) == 1:
				sys.stderr.write('...{} is exclusive chromosome name! Specify new name before deletion!\n'.format(name))
			else:
				del pangenome.chrom_names[name]
				chrom_names.remove(name)
				if chrom.name == name: chrom.name = chrom_names[0]
		except:
			sys.stderr.write('...unable to find chromosome name {} in {}! ...SKIPPING\n'.format(name, panfile))
	try:
		pangenome.compress(panfile)
	except:
		sys.stderr.write('Error while pangenome compressing\n')

def new_name(panfile, chrom_name, new_name):
	try:
		pangenome = decompress(panfile)
	except:
		raise ValueError('...unable to open {}!'.format(panfile))
		
	if chrom_name not in pangenome.chrom_names: 
		raise ValueError('Chromosome {} could not be found in PanCake Data Object from file {}!'.format(chrom_name, panfile))

	chrom = pangenome.chrom_names[chrom_name]
	genome_flag = chrom_name == chrom.genome
	chrom.new_name(new_name)
	pangenome.chrom_names[new_name] = chrom
	if genome_flag and chrom_name in pangenome.genomes and len(pangenome.genomes[chrom_name])==1:
		del pangenome.genomes[chrom_name]
		pangenome.genomes[new_name] = set()
		pangenome.genomes[new_name].add(chrom)
		chrom.genome = new_name
	try:
		pangenome.compress(panfile)
	except:
		sys.stderr.write('Error while pangenome compressing\n')

def add_to_genome(panfile, chroms, genome):
	try:
		pangenome = decompress(panfile)
	except:
		raise ValueError('...unable to open {}!'.format(panfile))
	for chrom_name in chroms:
		chrom = pangenome.chrom_names[chrom_name]
		pangenome.genomes[chrom.genome].remove(chrom)
		if len(pangenome.genomes[chrom.genome]) == 0: del pangenome.genomes[chrom.genome]
		chrom.genome = genome
		if genome not in pangenome.genomes:
			pangenome.genomes[genome] = set()
		pangenome.genomes[genome].add(chrom)
	try:
		pangenome.compress(panfile)
	except:
		sys.stderr.write('Error while pangenome compressing\n')

def include_genome_info(panfile, genome_file):
	'''specifyChrom: include information from genome file provided by the user'''
	try:
		pangenome = decompress(panfile)
	except:
		raise ValueError('...unable to open {}!'.format(genome_file))
	try:
		pangenome = decompress(panfile)
	except:
		raise ValueError('...unable to open {}!'.format(genome_file))
	with open(genome_file) as genomes:
		for line in genomes:
			ll = line.strip().split('\t')
			if len(ll)>1:
				genome = ll[0]
				chrom_name = ll[1]
				#change chromosome name
				if len(ll)==3: 
					new_name = ll[2]
					chrom = pangenome.chrom_names[chrom_name]
					genome_flag = True if chrom_name == chrom.genome else False
					chrom.new_name(new_name)
					pangenome.chrom_names[new_name] = chrom
					if genome_flag and chrom_name in pangenome.genomes and len(pangenome.genomes[chrom_name])==1:
						del pangenome.genomes[chrom_name]
						pangenome.genomes[new_name] = set()
						pangenome.genomes[new_name].add(chrom)
						chrom.genome = new_name
					chrom_name = new_name
				try:
					chrom = pangenome.chrom_names[chrom_name]
					pangenome.genomes[chrom.genome].remove(chrom)
					if len(pangenome.genomes[chrom.genome]) == 0: del pangenome.genomes[chrom.genome]
					chrom.genome = genome
					if genome not in pangenome.genomes:
						pangenome.genomes[genome] = set()
					pangenome.genomes[genome].add(chrom)
				except:
					sys.stderr.write('Chromosome {} not in pangenome!\n'.format(chrom_name))
	try:
		pangenome.compress(panfile)
	except:
		sys.stderr.write('Error while pangenome compressing\n')


def sequence(panfile, genome_name, chrom_name, start, stop, outfile='', lw=100):
	try:
		pangenome = decompress(panfile)
	except:
		raise ValueError('...unable to open {}!'.format(panfile))
	
	if chrom_name:
		try:
			chrom= pangenome.chrom_names[chrom_name] 
		except:
			raise ValueError('Chromosome {} could not be found in PanCake Object from file {}!'.format(chrom_name, panfile))
	
		chrom_length = pangenome.chrom_length(chrom)
		if start> chrom_length:
			raise ValueError('Specified START position {} > length of chromosome {} ({}bp)!'.format(start, chrom, chrom_length))

		if not stop:
			stop = chrom_length
		elif stop > chrom_length:
			raise ValueError('Specified STOP position > length of chromosome {} ({}bp)!'.format(stop, chrom, chrom_length))
	
		#get sequence
		seq = pangenome.seq(chrom,start, stop)
		if outfile:
			fasta_header = '>{} {}-{} ({}bp)\n'.format(chrom_name, start, stop, stop-start+1)
			seq='\n'.join(seq[i:i+lw] for i in range(0, len(seq), lw)) + '\n' #insert linebreaks
			dir, f = os.path.split(outfile)
			if dir:
				try:
					os.stat(dir)
				except:
					os.mkdir(dir)
		
			with open(outfile, 'wt') as fasta_file:
				fasta_file.write(fasta_header)
				fasta_file.write(seq)
		else:
			print(seq)
	elif genome_name:
		try:
			chroms = pangenome.genomes[genome_name]
		except:
			raise ValueError('Genome {} could not be found in PanCake Object from file {}!'.format(genome_name, panfile))
		chroms = sorted(list(chroms), key= lambda x: pangenome.chromosome_stops[x].stop, reverse=True)
		if outfile:
			dir, f = os.path.split(outfile)
			if dir:
				try:
					os.stat(dir)
				except:
					os.mkdir(dir)
			open(outfile, 'w').close()
		for chrom in chroms:
			stop = pangenome.chromosome_stops[chrom].stop
			fasta_header = '>{} {}-{} ({}bp)\n'.format(chrom.name, 1, stop, stop)
			seq = pangenome.seq(chrom)
			seq='\n'.join(seq[i:i+lw] for i in range(0, len(seq), lw)) + '\n' #insert linebreaks
			if outfile:
				with open(outfile, 'at') as fasta_file:
					fasta_file.write(fasta_header)
					fasta_file.write(seq)
			else:
				print(fasta_header)
				print(seq)



def initialize_gv_output(pan_file, chroms, starts, stops, all_flag, region_flag, max_entries, max_nodes, max_edges, output):

	#open panfile
	try:
		pangenome=decompress(pan_file)
	except:
		sys.stderr.write('#\n...unable to initialize pangenome from {}!\n'.format(pan_file))
		return
	sys.stderr.write('#\n...pangenome loaded from {}\n#\n'.format(pan_file))
	
	final_chroms, final_starts, final_stops = [], [], []
	
	if not chroms:
		sys.stderr.write('...no chromosomes specified ...taking all chromosomes from {}\n'.format(pan_file))
		final_chroms = list(pangenome.chromosome_stops.keys())
	else:
		for i in range(len(chroms)):
			try:
				chrom = pangenome.chrom_names[chroms[i]]
			except:
				raise ValueError('Chromosome {} could not be found in PanCake Object from file {}!'.format(chroms[i], pan_file))
			final_chroms.append(chrom)
	for i in range(len(final_chroms)):
		chrom_len = pangenome.chrom_length(final_chroms[i])
		#treat start
		if starts and i<len(starts) and starts[i]:
			if starts[i] < 1:
				sys.stderr.write('...START position {} on chromosome {} <1 ...setting to 1\n'.format(starts[i], final_chroms[i]))
				starts[i] = 1
			elif starts[i] > chrom_len:
				sys.stderr.write('...START position {} on chromosome {} > chromosome length {}bp ...setting to chromosome length\n'.format(starts[i], final_chroms[i], chrom_len))
				starts[i] = chrom_len
			final_starts.append(starts[i])
		else: final_starts.append(1)
		#treat stop
		if stops and i<len(stops) and stops[i]:
			if stops[i] < final_starts[i]:
				sys.stderr.write('...STOP position {} on chromosome {} < start position {} ...setting to chromosome length {}bp\n'.format(stops[i], final_chroms[i], final_starts[i], chrom_len))
				stops[i] = chrom_len
			elif stops[i] > chrom_len:
				sys.stderr.write('...STOP position {} on chromosome {} > chromosome length {}bp ...setting to chromosome length\n'.format(stops[i], final_chroms[i], chrom_len))
				stops[i] = chrom_len
			final_stops.append(stops[i])
		else: final_stops.append(chrom_len)

	sys.stderr.write('#\n...creating .gv file covering\n')
	for i in range(len(final_chroms)):
		sys.stderr.write('{}\t{}-{}\n'.format(final_chroms[i], final_starts[i], final_stops[i]))
	sys.stderr.write('#\n')
	get_gv(pangenome, final_chroms, final_starts, final_stops, all_flag,region_flag, max_entries, max_nodes, max_edges, output)



#DEBUG
#def printer(panfile):
#	try:
#		pangenome = decompress(panfile)
#	except:
#		raise ValueError('...unable to open {}!'.format(panfile))
#	print(pangenome)



import sys,os
from collections import defaultdict

####SINGLETONS

def get_genome_singletons(pancake, ref_genome, min_len, genomes, chromosomes, exclude_genomes, exclude_chromosomes, folder='', output=True, bed_file=''):
	#very much simplified in contrast to core search, just get the set of all chromosomes not allowed to be aligned with
	if ref_genome not in pancake.genomes:
		raise ValueError('Genome ' + str(ref_genome) + ' not in PanGenome!')
	else:
		ref_chroms = pancake.genomes[ref_genome]
	for chrom in ref_chroms:
		print('#\n...parsing {}'.format(chrom), file=sys.stderr)
		regions = get_singletons(pancake, chrom.name, min_len, genomes, chromosomes, exclude_genomes, exclude_chromosomes, folder, output, bed_file)
	return regions

def get_singletons(pancake, ref_chrom, min_len, chroms=[], genomes=[], exclude_genomes=[], exclude_chromosomes=[], folder='', output=True, bed_file=''):

	#initiation, error checking
	if ref_chrom not in pancake.chrom_names:
		raise ValueError('Chromosome ' + str(ref_chrom) + ' not in PanGenome!')
	else:
		ref_chrom =  pancake.chrom_names[ref_chrom]


	if output and not folder:
		folder=os.getcwd() + '/singletons_'+str(ref_chrom)
		try:
			os.stat(folder)
		except:
			os.mkdir(folder)
	elif output:
		try:
			os.stat(folder)
		except:
			os.mkdir(folder)

	#Here: get the set of all chromosomes not allowed to be aligned with ref_chrom
	all_chroms = set()
	if not genomes:
		for g in pancake.genomes:
			if g != ref_chrom.genome:
				for c in pancake.genomes[g]:
					all_chroms.add(c)
	else:
		for g in genomes:
			try:
				tmp_g = pancake.genomes[g]
				all_chroms = all_chroms.union(tmp_g)
			except:
				sys.stderr.write('...genome {} not in pangenome! ...SKIPPING\n'.format(g))

	for c in chroms:
		try:
			all_chroms.add(pancake.chrom_names[c])
		except:
			sys.stderr.write('...chromosome {} not in pangenome! ...SKIPPING\n'.format(c))
	if ref_chrom in all_chroms: all_chroms.remove(ref_chrom)

	for c in exclude_chromosomes:
		try:
			chrom = pancake.chrom_names[c]
			if chrom in all_chroms: all_chroms.remove(chrom)
		except:
			sys.stderr.write('...chromosome {} not in pangenome! ...SKIPPING\n'.format(c))
	for g in exclude_genomes:
		try:
			tmp_g = pancake.genomes[g]
			for c in tmp_g:
				if c in all_chroms: all_chroms.remove(c)
		except:
			sys.stderr.write('...genome {} not in pangenome! ...SKIPPING\n'.format(g))

	fi = pancake.chromosomes[ref_chrom][1] #current fi
	regions = list() #list of tuples
	min_length, max_length, m = 0,0,0
	with open(bed_file, 'a') as bedfile:
	#iterate over reference genome
		while fi:
			if not fi.sh_feat or fi.sh_feat.singleton_sf(all_chroms):
				if regions and regions[-1][1] +1 == fi.start:
					regions[-1] = (regions[-1][0], fi.stop)
				else:
					regions.append((fi.start, fi.stop))
			fi = fi.nextFI 
		count, l = 0, 0
		if regions:
			min_length, max_length = None, None
			for region in regions:
				region_length = region[1]-region[0]+1
				if region_length >= min_len:
					bedfile.write('{}\t{}\t{}\n'.format(ref_chrom, region[0], region[1]))
					if output:
						with open('{}/{}_{}-{}.fa'.format(folder, ref_chrom, region[0], region[1]), 'wt') as fasta:
							fasta.write('>{} {}-{} ({}bp)\n'.format(ref_chrom, region[0], region[1], region[1]-region[0]+1))
							fasta.write(pancake.seq(ref_chrom, region[0], region[1]) + '\n')
					count+=1
					l+=region_length
					if not min_length or region_length < min_length: min_length = region_length
					if not max_length or region_length > max_length: max_length = region_length
		sys.stderr.write('#\n')
		if count >0:
			print('Found {} singleton regions.\nMinimum Length = {}\nMaximum Length = {}\nMean = {}\n'.format(count, min_length, max_length, l/count))
			if output:
				print('Singleton regions written into folder {}'.format(folder))
		else:
			print('Found NO singleton regions.')



import sys,os
from collections import defaultdict
from pancake.utils import rev_comp

####CORE Identification

def get_genome_core(pancake, ref_genome, min_len, genomes, chroms, exclude_genomes, exclude_chromosomes, non_core_frac, nc_space, bed_file):
	#get all chromosomes of ref_genome
	if ref_genome not in pancake.genomes:
		raise ValueError('Genome ' + str(ref_genome) + ' not in PanGenome!')
	else:
		ref_chroms = set([x.name for x in pancake.genomes[ref_genome]])
	regions=dict()
	#compute excludes: all chromosomes of ref_genome
	exclude_chromosomes = set(exclude_chromosomes).union(ref_chroms)
	for chrom in ref_chroms:
		print('#\n...parsing {}'.format(chrom), file=sys.stderr)
		regions = get_core(pancake, chrom, min_len, genomes, chroms, exclude_genomes, exclude_chromosomes, non_core_frac, nc_space, regions, bed_file)
	return regions


#get core for 1 chromosome; called nchromosomes-times if ref_genome is specified
def get_core(pancake, ref_chrom, min_len, genomes=[], chromlist=[], exclude_genomes=[], exclude_chromosomes=[], non_core_frac=0.1, nc_space=25, regions=dict(), bed_file='core.bed'):

	#initiation, error checking
	if ref_chrom not in pancake.chrom_names:
		raise ValueError('Chromosome ' + str(ref_chrom) + ' not in PanGenome!')
	else:
		ref_chrom =  pancake.chrom_names[ref_chrom]
	####HERE deal with excludes
	if not genomes:
		genomes = set([x.genome for x in pancake.chromosomes])
	else:
		genomes = set(genomes)
	if ref_chrom.genome in genomes: genomes.remove(ref_chrom.genome)
	
	chroms = set()
	tmp_genomes = set()
	for chrom in chromlist: 
		if chrom not in pancake.chrom_names: 
			raise ValueError('Chromosome ' + str(chrom) + ' not in PanGenome!')
		chrom = pancake.chrom_names[chrom]
		chroms.add(chrom)
		if chrom.genome in genomes: 
			tmp_genomes.add(chrom.genome)
	genomes = genomes.difference(tmp_genomes)
	genomes = genomes.difference(set(exclude_genomes))

	exclude_dict = defaultdict(set)
	for chrom in exclude_chromosomes:
		if chrom not in pancake.chrom_names: 
			raise ValueError('Chromosome ' + str(chrom) + ' not in PanGenome!')
		chrom = pancake.chrom_names[chrom]
		exclude_dict[chrom.genome].add(chrom)
	for entry in exclude_dict:
		if entry in genomes: genomes.remove(entry)
		for c in pancake.genomes[entry].difference(exclude_dict[entry]):
			chroms.append(c)

	#OUTPUT 
	print('against', file=sys.stderr)
	if genomes:
		print('GENOMES', file=sys.stderr)
		for g in genomes:
			print(' ' + g, file=sys.stderr)
	if chroms:
		print('#\nCHROMOSOMES', file=sys.stderr)
		for c in chroms:
			print( ' ' + str(c), file=sys.stderr)
	print('#', file=sys.stderr)
	###END OUTPUT

	#identify core FIs and core SFs
	fi = pancake.chromosomes[ref_chrom][1]
	core_fis = []
	l=0
	core_sfs = []
	while fi:
		if fi.sh_feat: 
			if fi.sh_feat in core_sfs:
				core_fis.append(fi)
				l+=1
			elif core_sf(fi, genomes, chroms):
				core_fis.append(fi)
				core_sfs.append(fi.sh_feat)
				l+=1
		fi=fi.nextFI
	print('Found {} core FIs!\n#'.format(l), file=sys.stderr)


	regions[ref_chrom] = dict()
	starts=list()
	min_length, max_length, m = 0,0,0
	#iterate over reference genome
	fi_count, core_fi_count = 0, 0
	with open(bed_file, 'a') as bedfile:

		for i in range(len(core_fis)):
			if i != 0 and i%1000 == 0:
				print('...{} Feature Instances parsed (current chromosome position = {})'.format(i, core_fis[i].start))
			region = get_region(i, core_fis, pancake, genomes, chroms, non_core_frac, nc_space)
			fi = core_fis[i]
			#decide here if add to regions
			add_flag = False
			if not starts:
				add_flag =True
			elif starts[-1][1] != region[fi][-1].stop:
				add_flag = True
			ref_tupel = (fi.start, region[fi][-1].stop)
			if ref_tupel[1]-ref_tupel[0]+1 < min_len: add_flag = False
			if add_flag:
				bedfile.write('{}\t{}\t{}\n'.format(ref_chrom, ref_tupel[0], ref_tupel[1]))
				#add to regions (save for output) and include into statistics
				starts.append(ref_tupel)
				l = ref_tupel[1]-ref_tupel[0]+1
				if not min_length or l<min_length: min_length = l
				if not max_length or l>max_length: max_length = l
				m+=l
				regions[ref_chrom][ref_tupel] = dict()#dict of chromosomes (for sorted .fasta output)
				del region[fi]
				for entry in region:
					entry_list = region[entry] #list of core_sfs
					c = entry_list[0].chrom
					if c not in regions[ref_chrom][ref_tupel]: regions[ref_chrom][ref_tupel][c] = list()
					if not entry_list[0].seq.reverse:
						#seq in '+' direction
						regions[ref_chrom][ref_tupel][c].append((entry_list[0].start, entry_list[-1].stop, True))
					else:
						regions[ref_chrom][ref_tupel][c].append((entry_list[-1].start, entry_list[0].stop, False))
	
	if len(starts) >0:
		print('Found ' + str(len(starts)) + ' core regions.\nMinimum Length = '+str(min_length)+'\nMaximum Length = '+ str(max_length) + '\nMean = '+str(m/len(starts))+'\n')#, file=sys.stderr)
	else:
		print('Found NO core regions.')
	return regions


def core_sf(fi, genomes, chromosomes): 
	'''return True if all FIs from genome and chromosomes in fi.sf'''
	#genomes without ref_chrom.genome
	sf = fi.sh_feat
	entries = sf.entries
	l = len(entries)
	if genomes:
		if l < len(genomes):
			return False
		elif not genomes.issubset(set([x.chrom.genome for x in entries])):
			return False
	if chromosomes:
		if l < len(chromosomes):
			return False
		elif not chromosomes.issubset(set([x.chrom for x in entries])):
			return False
	return True


def get_region(j, core_fis, pancake, genomes, chroms, non_core_frac=0.1, nc_space=25):
	'''identifies longest possible core region starting at ref_fi'''
	ref_fi = core_fis[j]
	#initiation
	if ref_fi.seq.reverse: ref_fi.sh_feat.turn()
	region_dic = dict()

	all_chroms = chroms.union(set([ref_fi.chrom]))
	for g in genomes: 
		for c in pancake.genomes[g]:
			all_chroms.add(c)
	for fi in ref_fi.sh_feat:
		if fi.chrom in all_chroms:
			region_dic[fi] = [fi] #sequence of core fis for all chromosomes

	core_flag = True
	ref_space=0
	ref_length=ref_fi.stop-ref_fi.start+1
	#start iteration here
	#find next core_sf for ref_fi
	while core_flag:
		next_ref_core_fi = next_core(j, core_fis, non_core_frac, nc_space, space=ref_space,length=ref_length)
		if next_ref_core_fi:
			j+=1
			#next core_sf found
			ref_space += next_ref_core_fi.start - region_dic[ref_fi][-1].stop +1
			ref_length += next_ref_core_fi.stop - region_dic[ref_fi][-1].stop +1
			region_dic[ref_fi].append(next_ref_core_fi)
			next_core_sf = next_ref_core_fi.sh_feat
			for start_fi in region_dic:
				#iterate over non-reference fis
				if start_fi != ref_fi:
					space=0
					length = start_fi.stop-start_fi.start+1
					next_core_fi = None
					if not start_fi.seq.reverse:
						#go forward through pangenome
						next_fi = region_dic[start_fi][-1].nextFI
						while space/length < non_core_frac and next_fi and not next_core_fi and space <= nc_space:
							if next_fi.sh_feat and next_fi.sh_feat == next_core_sf and not next_fi.seq.reverse:
								next_core_fi = next_fi
							else:
								space += next_fi.stop -next_fi.start+1
								length += next_fi.stop -next_fi.start+1
							next_fi = next_fi.nextFI
					else:
						#go backward through pangenome
						next_fi = region_dic[start_fi][-1].prevFI
						while space/length < non_core_frac and next_fi and not next_core_fi and space <= nc_space:
							if next_fi.sh_feat and next_fi.sh_feat == next_core_sf and next_fi.seq.reverse:
								next_core_fi = next_fi
							else:
								space += next_fi.stop -next_fi.start+1
								length += next_fi.stop -next_fi.start+1
							next_fi = next_fi.prevFI
					region_dic[start_fi].append(next_core_fi)
		else: core_flag = False

		if core_flag==True:
			tmp_chroms = set()
			for fi in region_dic:
				if region_dic[fi][-1]: tmp_chroms.add(fi.chrom)

			#decide if chromosomes still represent a valid core, i.e. one chrom from each genome
			for g in genomes:
				if not pancake.genomes[g].intersection(tmp_chroms):
					core_flag=False
					break
			if core_flag:
				for c in chroms:
					if c not in tmp_chroms: core_flag=False
			
			#not a valid core region any more
			if core_flag == False:
				for fi in region_dic:
					region_dic[fi] = region_dic[fi][:-1]

			else:
				#still a valid core region, but some paths might be cutted
				fi_list = list(region_dic.keys())
				for fi in fi_list:
					if not region_dic[fi][-1]:
						del region_dic[fi]

	return region_dic


def next_core(j, core_fis, non_core_frac, nc_space, space=0,length=0):
	'''return next fi of a core sf fullfilling given constraints, or NONE if there is none'''
	if j+1 < len(core_fis):
		next_fi = core_fis[j+1]
		tmp_space = next_fi.start - core_fis[j].stop -1
		l=next_fi.stop -next_fi.start+1
		if tmp_space < nc_space and (space+tmp_space)/(length+tmp_space+l) <= non_core_frac:
			return next_fi


#CORE .fasta output
def write_core_fasta_files(pancake, regions, ref_genome = None, folder=None, iterator=False):

	if not folder: 
		if ref_genome:
			folder=os.getcwd() + 'core_'+str(ref_chrom)
		elif len(regions)==1:
			folder=os.getcwd() + 'core_'+str(list(regions.keys())[0])
		else:
			folder = os.getcwd() + 'core_regions'
	try:
		os.stat(folder)
	except:
		os.mkdir(folder)

	for ref_chrom in regions.keys():
		starts = list(regions[ref_chrom].keys())
		starts.sort(key = lambda item: item[0], reverse=True)#sort by start value
		other_chroms = list(set(pancake.chromosomes.keys()).difference(set([ref_chrom])))
		core_count =1

		for entry in starts:
			#write .fas file for 1 core region
			with open('{}/{}_{}-{}.fas'.format(folder,ref_chrom,entry[0],entry[1]), 'wt') as fasta:
				#write core sequence of reference chromosome
				fasta.write('>{} {}-{}|len={}bp|+\n'.format(ref_chrom, entry[0], entry[1], entry[1]-entry[0]+1) )
				fasta.write(pancake.seq(ref_chrom, entry[0], entry[1])+'\n')
				#write (possible) other subsequences of reference chromosome
				if ref_chrom in regions[ref_chrom][entry]:
					regions[ref_chrom][entry][ref_chrom].sort(key=lambda item: item[0], reverse=True)#sort by start value
					for region in regions[ref_chrom][entry][ref_chrom]:
						fasta_header = '>{} {}-{}|len={}bp|'.format(ref_chrom, region[0], region[1], region[1]-region[0]+1)
						if region[2]:
							fasta.write(fasta_header + '+\n' + pancake.seq(ref_chrom, region[0], region[1])+'\n')
						else:
							fasta.write(fasta_header + '-\n' + rev_comp(pancake.seq(ref_chrom, region[0], region[1])) +'\n')
				#write subsequences of other chroms
				for chrom in other_chroms:
					if chrom in regions[ref_chrom][entry]:
						regions[ref_chrom][entry][chrom].sort(key=lambda item: item[0], reverse=True)#sort by start value for one chromosome
						for region in regions[ref_chrom][entry][chrom]:
							fasta_header = '>{} {}-{}|len={}bp|'.format(chrom, region[0], region[1], region[1]-region[0]+1)
							if region[2]:
								fasta.write(fasta_header+'+\n')
								fasta.write(pancake.seq(chrom, region[0], region[1]))
							else:
								fasta.write(fasta_header + '-\n')
								fasta.write(rev_comp(pancake.seq(chrom, region[0], region[1])))
							fasta.write('\n')




import sys

class Alignment:

	def __init__(self, qname, sname, qtuple, stuple, alignment, qstrand=None, sstrand=None):
		self.qali = alignment[0].upper()
		self.qname = qname
		self.qstart, self.qstop, self.qstrand = get_ali_items(qtuple, qstrand)
		self.sali = alignment[1].upper()
		self.sname = sname
		self.sstart, self.sstop, self.sstrand = get_ali_items(stuple, sstrand)

	def __str__(self):
		out = '{}: ({}, {}) '.format(self.qname, self.qstart, self.qstop)
		out += '+\n' if self.qstrand else '-\n'
		out += '{}: ({}, {}) '.format(self.sname, self.sstart, self.sstop)
		out += '+\n' if self.sstrand else '-\n'
		out+= self.qali +'\n'
		out+= self.sali
		return out

	def convert(self):
		'''change query and subject'''
		self.qali, self.sali = self.sali, self.qali
		self.qname, self.sname = self.sname, self.qname
		self.qstart, self.sstart = self.sstart, self.qstart
		self.qstop, self.sstop = self.sstop, self.qstop
		self.qstrand, self.sstrand = self.sstrand, self.qstrand


	#q indicates if cut_ind correspond to query or subject
	#cut_ind is always from alignment start (neither its + or -)
	def divide(self, cut_ind, q=True):
		ref_ali, not_ref_ali = (self.qali, self.sali) if q else (self.sali, self.qali)
		ref_len, not_ref_len, i = 0, 0, 0

		while ref_len<cut_ind:
			#only non-deletions count to ref_len and not_ref_len
			ref_len += ref_ali[i] != '-'
			not_ref_len += not_ref_ali[i] != '-' 
			i+=1
		new_alignment = (self.qali[:i], self.sali[:i])
		#just changing ref_len and not_ref_len once (for speed reasons)
		if not q: ref_len, not_ref_len = not_ref_len, ref_len
		qtuple1, qtuple2 = None, None
		if ref_len>0 and not_ref_len>0:
			qtuple1 = (self.qstart, self.qstart+ref_len-1) if self.qstrand else (self.qstop-ref_len+1, self.qstop)
			stuple1 = (self.sstart, self.sstart+not_ref_len-1) if self.sstrand else (self.sstop-not_ref_len+1, self.sstop)
		if bool(filter(lambda x: x!= '-',self.qali[i:] )) and bool(filter(lambda x: x!= '-',self.sali[i:] )):
			qtuple2 = (self.qstart+ref_len, self.qstop) if self.qstrand else (self.qstart, self.qstop-ref_len)
			stuple2 = (self.sstart+not_ref_len, self.sstop) if self.sstrand else (self.sstart, self.sstop-not_ref_len)

		if qtuple1:
			new_ali = Alignment(self.qname, self.sname, qtuple1, stuple1, new_alignment, qstrand=self.qstrand, sstrand=self.sstrand)
		if qtuple2:
			self.qali, self.sali = self.qali[i:], self.sali[i:]
			self.qstart, self.qstop = qtuple2[0], qtuple2[1]
			self.sstart, self.sstop = stuple2[0], stuple2[1]

		if qtuple1 and qtuple2:
			return new_ali, self
		elif qtuple1:
			return new_ali, None
		elif qtuple2:
			return None, self
		else: return None, None


def get_ali_items(tuple, strand=None):
	if strand == None:
		strand = tuple[0] <= tuple[1] #True, if forward
	return (min(tuple), max(tuple), strand)




import numpy
#from editops import *

#scores for needle wunsch alignment
match_score = 10
mm_score = -5
gap_score = -5 #for opening and extending 


###

def rev_comp(s):
	return s.translate(str.maketrans('ACGTacgtRYMKrymkVBHDvbhd', 'TGCAtgcaYRKMyrkmBVDHbvdh'))[::-1]


def delete_gaps(seq):
	return seq.replace('-','')

####alignment modifications

def needle_wunsch(s1, s2):

	m,n = len(s1),len(s2)
	table = numpy.zeros((m+1, n+1)) #DP table

	# fill table
	for i in range(0, m+1): table[i][0] = gap_score * i
	for j in range(0, n+1): table[0][j] = gap_score * j
	for i in range(1, m+1):
		for j in range(1, n+1):
			match = table[i-1][j-1] + score(s1[i-1], s2[j-1])
			delete = table[i-1][j] + gap_score
			insert = table[i][j-1] + gap_score
			table[i][j] = max(match, delete, insert)

	#traceback
	ali1, ali2 = '', ''
	i,j = m,n # start from bottom right
	while i > 0 and j > 0:
		current = table[i][j]
		diagonal = table[i-1][j-1]
		up = table[i][j-1]
		left = table[i-1][j]

		if current == diagonal + score(s1[i-1], s2[j-1]):
			ali1 += s1[i-1]
			ali2 += s2[j-1]
			i -= 1
			j -= 1
		elif current == left + gap_score:
			ali1 += s1[i-1]
			ali2 += '-'
			i -= 1
		elif current == up + gap_score:
			ali1 += '-'
			ali2 += s2[j-1]
			j -= 1

	#continue to top left cell
	while i > 0:
		ali1 += s1[i-1]
		ali2 += '-'
		i -= 1
	while j > 0:
		ali1 += '-'
		ali2 += s2[j-1]
		j -= 1
	return(ali1[::-1], ali2[::-1])


def score(b1, b2):
	if b1 == b2: 
		return match_score
	elif b1 == '-' or b1 == '-':
		return gap_score
	else:
		return mm_score


def get_new_alignment(ali1, ali2):
	#return (heuristic) alignment of 2nd sequence in ali1 and 2nd sequence in ali2
	#returns only 'gapped' sequences, not an Alignment instance
	ali1_ind = 0
	ali2_ind = 0
	new_alignment = [ [], [] ]
	lali1, lali2 = len(ali1[0]), len(ali2[0])
	ali1_0, ali1_1, ali2_0, ali2_1 = ali1[0], ali1[1], ali2[0], ali2[1]
	while ali1_ind < lali1 and ali2_ind < lali2:
		if ali1_0[ali1_ind] != '-':
			if ali2_0[ali2_ind] != '-':
				#prevent "double gaps"
				if ali1_1[ali1_ind] != '-' or ali2_1[ali2_ind] != '-':
					new_alignment[0].append(ali1_1[ali1_ind])
					new_alignment[1].append(ali2_1[ali2_ind])
				ali1_ind += 1
				ali2_ind += 1
			else:
				# gap in ali2_0
				new_alignment[0].append('-')
				new_alignment[1].append(ali2_1[ali2_ind])
				ali2_ind += 1
		else:
			# gap in ali1_0
			if ali2_0[ali2_ind] != '-':
				new_alignment[0].append(ali1_1[ali1_ind])
				new_alignment[1].append('-')
				ali1_ind += 1
			else:
				#gap in ali2_0
				#prevent "double gaps"
				if ali1_1[ali1_ind] != '-' or ali2_1[ali2_ind] != '-':
					new_alignment[0].append(ali1_1[ali1_ind])
					new_alignment[1].append(ali2_1[ali2_ind])
				ali1_ind += 1
				ali2_ind += 1

	a1 = ''.join(new_alignment[0])
	a2 = ''.join(new_alignment[1])
	
	#remaining bases?
	if ali1_ind < lali1: 
		a1 += ali1_1[ali1_ind:]
		a2 += '-' * (lali1 - ali1_ind)
	if ali2_ind < lali2: 
		a2 += ali2_1[ali2_ind:]
		a1 += '-' * ( lali2 - ali2_ind)

	return (a1, a2 )


def ali_rev_comp(ali):
	#get reverse complement of 2 aligned sequences
	new1 = ali[0].translate(str.maketrans('ACGTacgtRYMKrymkVBHDvbhd', 'TGCAtgcaYRKMyrkmBVDHbvdh'))[::-1]
	new2 = ali[1].translate(str.maketrans('ACGTacgtRYMKrymkVBHDvbhd', 'TGCAtgcaYRKMyrkmBVDHbvdh'))[::-1]
	return (new1, new2)

def new_ops_from_ops(op1,op2,ref):
	#op1,op2=editops corresponding to ref
	#return new representaion of op2 corresponding to op1
	ref_ind, op1_ind, op2_ind = 0,0,0
	new_op2 = []
	op2_match, op2_del = 0,0
	while op1_ind < len(op1) and op2_ind<len(op2):
		if type(op1[op1_ind]) == int:
			if op1[op1_ind] > 0:
				#match
				if type(op2[op2_ind]) == int:
					if op2[op2_ind]>0:
						#match,match --> match in new_op2
						if op2_del: 
							new_op2.append(-op2_del)
							op2_del = 0
						match_len = min(op1[op1_ind], op2[op2_ind])
						ref_ind += match_len
						op2_match += match_len
						if op1[op1_ind] > match_len:
							op1[op1_ind] -= match_len
						else: op1_ind +=1
						if op2[op2_ind] > match_len:
							op2[op2_ind] -= match_len
						else: op2_ind +=1
					elif op2[op2_ind]<0:
						#match,del --> deletion in new_op2
						if op2_match: 
							new_op2.append(op2_match)
							op2_match = 0
						del_len = min(op1[op1_ind], -op2[op2_ind])
						ref_ind += del_len
						op2_del += del_len
						if op1[op1_ind] > del_len:
							op1[op1_ind] -= del_len
						else: op1_ind +=1
						if -op2[op2_ind] > del_len:
							op2[op2_ind] += del_len
						else: op2_ind +=1
				else:
					if op2_del: 
						new_op2.append(-op2_del)
						op2_del = 0
					elif op2_match:
						new_op2.append(op2_match)
						op2_match = 0
					if op2[op2_ind].isupper():
						#match, substitution --> substitution in new_op2
						ref_ind +=1
						if op1[op1_ind] >1:
							op1[op1_ind] -= 1
						else: op1_ind +=1
						new_op2.append(op2[op2_ind])
						op2_ind+=1
					elif op2[op2_ind].islower():
						#match,insertion --> insertion in new_op2
						new_op2.append(op2[op2_ind])
						op2_ind+=1
			else:
				#deletion
				if type(op2[op2_ind]) == int:
					if op2[op2_ind]>0:
						#deletion,match --> insertion in new_op2
						#
						if op2_del: 
							new_op2.append(-op2_del)
							op2_del = 0
						elif op2_match:
							new_op2.append(op2_match)
							op2_match = 0
						#
						del_length = min(-op1[op1_ind],op2[op2_ind])
						if -op1[op1_ind] > del_length:
							op1[op1_ind] += del_length
						else: op1_ind+=1
						for b in ref[ref_ind:ref_ind+del_length]:
							new_op2.append(b.lower())
						ref_ind += del_length
						if op2[op2_ind] > del_length:
							op2[op2_ind] -= del_length
						else:
							op2_ind+=1
					else:
						#deletion,deletion --> just skip
						del_length = min(-op1[op1_ind],-op2[op2_ind])
						if -op1[op1_ind] > del_length:
							op1[op1_ind] += del_length
						else: op1_ind+=1
						if -op2[op2_ind] > del_length:
							op2[op2_ind] += del_length
						else: op2_ind+=1
						ref_ind+=del_length
				else:
					#deletion, substitution --> insertion
					#deletion, insertion --> insertion
					if op2_del: 
						new_op2.append(-op2_del)
						op2_del = 0
					elif op2_match:
						new_op2.append(op2_match)
						op2_match = 0
						
					if op2[op2_ind].isupper():
						#deletion, substitution --> insertion
						if -op1[op1_ind]>1:
							op1[op1_ind] += 1
						else: op1_ind +=1
						ref_ind+=1
					new_op2.append(op2[op2_ind].lower())
					op2_ind+=1
					
		elif op1[op1_ind].isupper():
			#substitution
			if type(op2[op2_ind]) == int:
				if op2[op2_ind]>0:
					#subtitution, match --> substitution in op2
					if op2_del: 
						new_op2.append(-op2_del)
						op2_del = 0
					elif op2_match:
						new_op2.append(op2_match)
						op2_match = 0
					new_op2.append(ref[ref_ind])
					if op2[op2_ind] > 1:
						op2[op2_ind] -=1
					else: op2_ind+=1
					ref_ind+=1
					op1_ind+=1
				else:
					#substitution, deletion --> deletion in op2
					if op2_match:
						new_op2.append(op2_match)
						op2_match = 0
					if -op2[op2_ind] > 1:
						op2[op2_ind]+=1
					else: op2_ind+=1
					ref_ind+=1
					op1_ind+=1
					op2_del+=1
			elif op2[op2_ind].isupper():
				#substitution, substitution --> match or substitution
				if op2_del: 
					new_op2.append(-op2_del)
					op2_del = 0
				#
				if op1[op1_ind]==op2[op2_ind]:
					op2_match+=1
				else:
					if op2_match:
						new_op2.append(op2_match)
						op2_match = 0
					new_op2.append(op2[op2_ind])
				op1_ind+=1
				op2_ind+=1
				ref_ind+=1
			else:
				#substitution,insertion --> insertion
				if op2_del: 
					new_op2.append(-op2_del)
					op2_del = 0
				elif op2_match:
					new_op2.append(op2_match)
					op2_match = 0
				#
				new_op2.append(op2[op2_ind])
				op2_ind+=1
		else:
			#insertion
			if type(op2[op2_ind]) == int:
				if op2[op2_ind]>0:
					#insertion, match --> deletion in op2
					if op2_match:
						new_op2.append(op2_match)
						op2_match = 0
					op2_del+=1
				else:
					#insertion, deletion -->  deletion in op2
					if op2_match:
						new_op2.append(op2_match)
						op2_match = 0
					op2_del+=1
				op1_ind+=1
			elif op2[op2_ind].isupper():
				#insertion, substitution --> deletion #TODO think about
				if op2_match:
					new_op2.append(op2_match)
					op2_match = 0
				op2_del +=1
				op1_ind+=1
			else:
				#insertion, insertion --> match or substitution
				if op2_del: 
					new_op2.append(-op2_del)
					op2_del = 0
				if op1[op1_ind]==op2[op2_ind]:
					#match
					op2_match+=1
				else:
					#substitution
					if op2_match:
						new_op2.append(op2_match)
						op2_match = 0
					#
					new_op2.append(op2[op2_ind].upper())
				op1_ind+=1
				op2_ind+=1

	
	if op2_match:
		new_op2.append(op2_match)
		op2_match = 0
		
	#treat remaining ops in op1 (insertions)
	for op in op1[op1_ind:]:
		op2_del += 1
	if op2_del: new_op2.append(-op2_del)

	#remaining ops in op2 are exclusively insertions (only present if no remaining ops in op1)
	for op in op2[op2_ind:]:
		new_op2.append(op)


	remaining_dels = len(ref[ref_ind:])
	if type(new_op2[-1]) == int and new_op2[-1] < 0:
			new_op2[-1] -= remaining_dels
	elif remaining_dels>0:
		new_op2 += [-remaining_dels]

	#TODO check while building
	i =1
	while i<len(new_op2):
		if (type(new_op2[i-1]), type(new_op2[i])) == (int,int) and new_op2[i-1]>0 and new_op2[i]>0:
			new_op2[i-1] += new_op2[i]
			new_op2 = new_op2[:i] + new_op2[i+1:]
		else: i+=1

	return new_op2


def ali_refinement(seq,ops):
	#align regions between matches correctly by needle_wunsch
	i = 0
	seq_ind, seq_l = 0, 0
	prev_op_ind = 0
	while i < len(ops):
		if type(ops[i]) == int:
			if ops[i]>0:
				#match
				if i >0:
					tmp_ops = ops[prev_op_ind:i]
					tmp_ali = ali_from_ops(tmp_ops, seq[seq_ind:seq_ind+seq_l])
					new_ali = needle_wunsch(tmp_ali[0].replace('-',''), tmp_ali[1].replace('-',''))
					new_tmp_ops = ops_from_ali(new_ali[0], new_ali[1])
					seq_ind += seq_l+ops[i]
					seq_l=0
					if prev_op_ind>0 and new_tmp_ops and ( type(ops[prev_op_ind-1]), type(new_tmp_ops[0])) == (int, int) and ops[prev_op_ind-1] >0 and new_tmp_ops[0] >0:
						ops[prev_op_ind-1] += new_tmp_ops[0]
						new_tmp_ops = new_tmp_ops[1:]
					if i < len(ops)-1 and new_tmp_ops and (type(ops[i]), type(new_tmp_ops[-1]))== (int,int) and ops[i] >0 and new_tmp_ops[-1]>0:
						ops[i] += new_tmp_ops[-1]
						new_tmp_ops = new_tmp_ops[:-1]
					ops = ops[:prev_op_ind] + new_tmp_ops + ops[i:] if prev_op_ind >0 else new_tmp_ops + ops[i:]
					#original
					prev_op_ind = prev_op_ind + len(new_tmp_ops) +1
					i= prev_op_ind-1 #is incremented at END of WHILE
					if len(new_tmp_ops)==0 and prev_op_ind>1:
						#take care of
						#AC-A
						#A-CA ---> three subsequent matches
						if type(ops[prev_op_ind-2])==int and ops[prev_op_ind-2] >0: # and type(ops[prev_op_ind-1])==int and ops[prev_op_ind-1] >0:
							#ops[prev_op_ind-1] is already known to be a match?
							ops[prev_op_ind-2] += ops[prev_op_ind-1]
							ops = ops[:prev_op_ind-1] + ops[prev_op_ind:]
							prev_op_ind -=1
							i-=1
				else: 
					prev_op_ind+=1
					seq_ind = ops[i]

			else:
				#deletion
				seq_l += abs(ops[i])
		elif ops[i].isupper():
			#substitution
			seq_l+=1
		else:
			#insertion
			pass
		i+=1
	#final
	if prev_op_ind < i-1:
		#print('final')
		tmp_ops = ops[prev_op_ind:]
		tmp_ali = ali_from_ops(tmp_ops, seq[seq_ind:])
		new_ali = needle_wunsch(tmp_ali[0].replace('-',''), tmp_ali[1].replace('-',''))
		new_tmp_ops = ops_from_ali(new_ali[0], new_ali[1])
		ops = ops[:prev_op_ind] + new_tmp_ops
	return ops


def ali_from_ops(ops, ref_seq):
	'''return alignment'''
	ref_ali_seq = ''
	ali_seq = ''
	seq_ind = 0
	for item in ops:
		if type(item) == int:
			if item >0:
				#matches
				ref_ali_seq += ref_seq[seq_ind:seq_ind+item]
				ali_seq += ref_seq[seq_ind:seq_ind+item]
				seq_ind += item
			else:
				#deletions
				ref_ali_seq += ref_seq[seq_ind:seq_ind+abs(item)]
				ali_seq += abs(item) * '-'
				seq_ind += -item
		elif item.islower():
			#insertion
			ref_ali_seq += '-'
			ali_seq += item.upper()
		else:
			#substitution
			ref_ali_seq += ref_seq[seq_ind]
			ali_seq += item.upper()
			seq_ind += 1
	return ref_ali_seq, ali_seq

def ops_from_ali(q_string, s_string):
	'''return EditOps ops computed from given alignment, ops represents sequence in s_string'''
	'''s_string is assumed to be already reverse complement, if reverse_flag set'''
	matches, deletions, o = 0, 0, []
	for qs, ss in zip(q_string, s_string):
		if qs.upper() == ss.upper():
			matches+=1
			if deletions>0: 
				o.append(-deletions)
				deletions=0
		else:
			if matches>0:
				o.append(matches)
				matches=0
			if ss == '-':
				deletions+=1
			else:
				if deletions>0:
					o.append(-deletions)
					deletions=0
				if qs == '-':
					o.append(ss.lower())
				else:
					o.append(ss.upper())
	if matches>0:
		o.append(matches)
	if deletions >0: o.append(-deletions)
	return o


def ops_from_ops_and_ali(ops, seq, qstring,sstring):
	#ops == editops for rebasing qstring to seq
	#return new_ops rebasing sequence of sstring to seq
	new_ops, ops_ind = [], 0
	seq_ind = 0
	matchs, dels=0,0
	#treat preceding deletions
	if type(ops[ops_ind]) == int and ops[ops_ind] < 0:
		dels += -ops[ops_ind]
		seq_ind+=-ops[ops_ind]
		ops_ind+=1
	for qs,ss in zip(qstring, sstring):
		if ops_ind < len(ops):
			if qs!= '-' and ss == '-':
				#deletion in ref_ali
				if type(ops[ops_ind]) == int:
					if ops[ops_ind]>0:
						#deletion in ref_ali, match in ops --> deletion in new_ops
						if matchs:
							new_ops.append(matchs)
							matchs=0
						dels+=1
						if ops[ops_ind]>1:
							ops[ops_ind]-=1
						else: ops_ind+=1
						seq_ind+=1
					else:
						pass #Should not happen
				elif ops[ops_ind].islower():
					#deletion in ref_ali, insertion in ops --> skip
					ops_ind+=1
				else:
					#deletion in ref_ali, substituion in ops --> deletion in new_ops
					if matchs:
						new_ops.append(matchs)
						matchs=0
					dels+=1
					#divide case for increment of seq_ind
					seq_ind+=1
					ops_ind+=1
			elif qs== '-' and ss != '-':
				#insertion in ref_ali --> always insertion in new_ops
				if dels:
					new_ops.append(-dels)
					dels=0
				elif matchs:
					new_ops.append(matchs)
					matchs=0
				new_ops.append(ss.lower())
			elif qs == ss:
				#match in ref_ali
				if type(ops[ops_ind]) == int:
					if ops[ops_ind]>0:
						#match in ref_ali, match in ops --> match in new_ops
						if dels:
							new_ops.append(-dels)
							dels=0
						matchs+=1
						seq_ind+=1
						if ops[ops_ind]>1:
							ops[ops_ind]-=1
						else: ops_ind+=1
					#else: #There should never be deletions
						#pass #TODO
						#match in ref_ali, deletion ops
				else:
					if dels:
						new_ops.append(-dels)
						dels=0
					elif matchs:
						new_ops.append(matchs)
						matchs=0
					if ops[ops_ind].islower():
						##match in ref_ali, insertion in ops --> insertion in new_ops
						new_ops.append(ss.lower())
						ops_ind+=1
					else:
						##match in ref_ali, substitution in ops --> substituion in new_ops
						new_ops.append(ss)
						ops_ind+=1
						seq_ind+=1
			else:
				#substituion in ref_ali
				if type(ops[ops_ind]) == int:
					if ops[ops_ind]>0:
						#substituion in ref_ali, match in ops --> substituion in new_ops
						if dels:
							new_ops.append(-dels)
							dels=0
						elif matchs:
							new_ops.append(matchs)
							matchs=0
						new_ops.append(ss)
						if ops[ops_ind]>1:
							ops[ops_ind]-=1
						else: ops_ind+=1
						seq_ind+=1
					else: pass
						#TODO should never be the case
				elif ops[ops_ind].islower():
					#substituion in ref_ali, insertion in ops --> insertion in new_ops
					if dels:
						new_ops.append(-dels)
						dels=0
					elif matchs:
						new_ops.append(matchs)
						matchs=0
					new_ops.append(ss.lower())
					ops_ind+=1
				else:
					#substituion in ref_ali, substituion in ops --> match or substitution in new_ops
					if dels:
						new_ops.append(-dels)
						dels=0
					if ss==seq[seq_ind]:
						#match
						matchs+=1
					else:
						#substitution
						if matchs:
							new_ops.append(matchs)
							matchs=0
						new_ops.append(ss)
					seq_ind+=1
					ops_ind+=1
			#TODO maybe treat 'double-gaps'
			#treat deletions in ops
			if ops_ind<len(ops) and type(ops[ops_ind])==int and ops[ops_ind]<0:
				if matchs:
					new_ops.append(matchs)
					matchs=0
				dels+= -ops[ops_ind]
				seq_ind += -ops[ops_ind]
				ops_ind+=1
		else:
			#ops closed
			#remaining ali should contain insertions exclusively
			if matchs:
				new_ops.append(matchs)
				matchs=0
			elif dels:
				new_ops.append(-dels)
				dels=0
			new_ops.append(ss.lower())
	####END ZIP
	
	if matchs:
		new_ops.append(matchs)
	elif dels:
		new_ops.append(-dels)
	return new_ops



def turn_ops(seq, ops):
	#return ops rebasing sequence of seq to the sequence given by ops
	seq_ind=0
	ops_ind=0
	new_ops = []
	matchs, dels = 0, 0
	for op in ops:
		if type(op)==int:
			if op>0:
				#match --> match
				if dels:
					new_ops.append(-dels)
					dels=0
				new_ops.append(op)
				seq_ind+=op
			else:
				#deletion --> insertion
				if dels:
					new_ops.append(-dels)
					dels=0
				for s in seq[seq_ind:seq_ind+(-op)]: new_ops.append(s.lower())
				seq_ind+=-op
		elif op.islower():
			#insertion --> deletion
			dels+=1
		else:
			#substituion --> substituion
			if dels:
				new_ops.append(-dels)
				dels=0
			new_ops.append(seq[seq_ind])
			seq_ind+=1
	if dels:
		new_ops.append(-dels)
	return new_ops


if __name__ == "__main__":
	pass




import sys
from pancake.alignment import Alignment
from pancake.utils import rev_comp
from pancake.pangenome import PanGenome


def parse_alis(f, pangenome, min_length, self_alignments=True):
	line_ind = 0
	with open(f, 'rt') as ali_file:
		for line in ali_file:
			if line[:6] == 'NUCMER':
				try:
					pangenome = parse_DELTA(f, pangenome, min_length, self_alignments)
				except:
					sys.stderr.write('ERROR while parsing {} ...SKIPPING\n'.format(f))
				break
			elif line[:5] == 'BLAST':
				try:
					pangenome = parse_BLAST(f, pangenome, min_length, self_alignments)
				except:
					sys.stderr.write('ERROR while parsing {} ...SKIPPING\n'.format(f))
				break
			if line_ind > 2:
				sys.stderr.write('...can not identify file format of {} ...SKIPPING\n'.format(f))
				break
			line_ind+=1
	return pangenome


#parse nucmer output, i.e. .delta files
#every Sequence in .delta file has to be already in pangenome
#Delta = (1, -3, 4, 0)
#A = ABC.DACBDCAC$
#B = .BCCDAC.DCAC$
def parse_DELTA(delta_file, pangenome, min_length, self_alignments=True):
	ali_count = 0
	with open(delta_file, 'r') as in_file:
		sys.stderr.write('...parsing .delta file {}\n'.format(delta_file))
		genome1, genome2 = '',''
		seq1, seq2 = '',''
		qtupel, stupel = None,None
		for line in in_file:
			if line[0]=='>':
				line_list = line.rstrip().split()
				genome1 = line_list[0][1:] # get rid of '>'
				genome2 = line_list[1]
			elif line[0] == '0': 
				if qtupel and stupel:
					#end of a pairwise alignment --> convert to Alignment Object
					ali = ( ali1 + seq1[ind1:].upper() , ali2 + seq2[ind2:].upper() ) # get gapped sequences
					ali_count+=1
					if (ali_count)%1000==0:
						pangenome.refine_unlabeled()
						pangenome.refine_adjacent_sfs()
						sys.stderr.write('...adding alignment {} ({} SFs created)\n'.format(ali_count, len(pangenome.features)))
					ali = Alignment(genome1, genome2, qtupel, stupel, ali)
					if ali.qname == ali.sname:
						#check self overlaps
						if ali.qstop > ali.sstart and ali.qstop < ali.sstop:
							#TODO
							ali_str = str(ali).split()[:2]
							sys.stderr.write('...excluding alignment\n\t{}\n\t{}\n'.format(ali_str[0], ali_str[1]))
							ali_count-=1
						elif ali.sstart > ali.qstart and ali.sstart< ali.qstop:
							#TODO
							ali_str = str(ali).split()[:2]
							sys.stderr.write('...excluding alignment\n\t{}\n\t{}\n'.format(ali_str[0], ali_str[1]))
							ali_count-=1
						else: pangenome.add_alignment(ali)
					else: pangenome.add_alignment(ali)
					qtupel, stupel = None, None
			elif genome1: #checking for genome1 makes sure not to parse header
				line_list = line.split(' ')
				if len(line_list)>1:
					if self_alignments or genome1 != genome2:
						if (genome1 != genome2 or line_list[0] != line_list[2] or line_list[1] != line_list[3]): #avoid whole genome alignment against itself
							#tuples
							qtupel = (int(line_list[0]), int(line_list[1]))
							stupel = (int(line_list[2]), int(line_list[3]))
							if qtupel[1]-qtupel[0]+1 >= min_length and max(stupel[1],stupel[0])-min(stupel[1],stupel[0])+1 >= min_length :
								ali1, ali2 = '', ''
								chrom1 = pangenome.chrom_names[genome1]
								chrom2 = pangenome.chrom_names[genome2]
								seq1 = pangenome.seq(chrom1, qtupel[0], qtupel[1]).upper() if qtupel[0]<qtupel[1] else rev_comp(pangenome.seq(chrom1, qtupel[1], qtupel[0])).upper()
								seq2 = pangenome.seq(chrom2, stupel[0], stupel[1]).upper() if stupel[0]<stupel[1] else rev_comp(pangenome.seq(chrom2, stupel[1], stupel[0])).upper()
								ind1, ind2 = 0, 0
							else: qtupel, stupel = None, None
				elif qtupel:
					#alignment modification
					if line[0] == '-':
						#deletion
						l = int(line[1:].rstrip())-1
						ali1 += seq1[ind1:ind1+l].upper() + '-'
						ali2 += seq2[ind2:ind2+l+1].upper()
						ind1 += l
						ind2 += l+1
						
					elif line[0] != '\n' and line[0] != '0':
						#insertion
						l=int(line.rstrip())-1
						ali1 +=seq1[ind1:ind1+l+1].upper()
						ali2 += seq2[ind2:ind2+l].upper() + '-'
						ind1 += l+1
						ind2 += l
	sys.stderr.write('...found {} valid pairwise alignments\n'.format(ali_count))
	return pangenome


def parse_BLAST(blast_file, pangenome, min_ali_length, self_alignments=True):
	''' parse a blastfile and return a customized data structure
	(q)uery - and (s)ubject keys : start, stop, seq'''
	new_dict = {} #stores current alignment
	new_dict['q'] = {"start" : [] , "stop" : [], "seq" : [] }
	new_dict['s'] = {"start" : [] , "stop" : [], "seq" : [] }
	with open(blast_file) as in_file:
		sys.stderr.write('...parsing BLAST file {}\n'.format(blast_file))
		ali_ind = 0 # if set: alignment line, gives first alignment index
		sbjct = ''
		for line in in_file:
			if line[:7] == 'Query= ': 
				query = line.split()[1]
			if line[0] == '>':
				if new_dict['q']['start'] and sbjct and (self_alignments or query!=sbjct):
					q_pos_tuple = (new_dict['q']['start'][0],new_dict['q']["stop"][-1] )
					s_pos_tuple = (new_dict['s']['start'][0], new_dict['s']['stop'][-1])
					if max(q_pos_tuple) - min(q_pos_tuple) +1 >= min_ali_length and max(s_pos_tuple) - min(s_pos_tuple) +1 >= min_ali_length:
						if query != sbjct or (min(q_pos_tuple), max(q_pos_tuple)) != (min(s_pos_tuple), max(s_pos_tuple)):
							ali_ind +=1
							if (ali_ind+1)%1000==0:
								pangenome.refine_unlabeled()
								pangenome.refine_adjacent_sfs()
								sys.stderr.write('...adding alignment {} ({} SFs created)\n'.format(ali_count, len(pangenome.features)))
							ali = Alignment(query, sbjct, q_pos_tuple, s_pos_tuple, (''.join(new_dict['q']['seq']), ''.join(new_dict['s']['seq'])))
							if ali.qname == ali.sname:
								#check self overlaps
								if ali.qstop > ali.sstart and ali.qstop < ali.sstop:
									#TODO
									ali_str = str(ali).split()[:2]
									sys.stderr.write('...excluding alignment\n\t{}\n\t{}\n'.format(ali_str[0], ali_str[1]))
									ali_count-=1
								elif ali.sstart > ali.qstart and ali.sstart< ali.qstop:
									#TODO
									ali_str = str(ali).split()[:2]
									sys.stderr.write('...excluding alignment\n\t{}\n\t{}\n'.format(ali_str[0], ali_str[1]))
									ali_count-=1
								else: pangenome.add_alignment(ali)
							else: pangenome.add_alignment(ali)
					new_dict['q'] = {"start" : [] , "stop" : [], "seq" : [] }
					new_dict['s'] = {"start" : [] , "stop" : [], "seq" : [] }
				sbjct = line.split()[0][1:] if line[1] != ' ' else line.split()[1]
			line_list = line.split()
			if line_list and line_list[0]=='Score' and 'Expect' in line_list:
				if new_dict['q']['start'] and (self_alignments or query!=sbjct): 
					q_pos_tuple = (new_dict["q"]["start"][0],new_dict["q"]["stop"][-1] )
					s_pos_tuple = (new_dict["s"]["start"][0], new_dict["s"]["stop"][-1])
					if max(q_pos_tuple) - min(q_pos_tuple) +1 >= min_ali_length and max(s_pos_tuple) - min(s_pos_tuple) +1 >= min_ali_length:
						if query != sbjct or self_alignments:
							if (min(q_pos_tuple), max(q_pos_tuple)) != (min(s_pos_tuple), max(s_pos_tuple)):
								ali_ind +=1
								if (ali_ind+1)%1000==0:
									sys.stderr.write('...adding alignment {} ({} SFs created)\n'.format(ali_count, len(pangenome.features)))
								ali = Alignment(query, sbjct, q_pos_tuple, s_pos_tuple, (''.join(new_dict["q"]["seq"]), ''.join(new_dict["s"]["seq"])))
								if ali.qname == ali.sname:
								#check self overlaps
									if ali.qstop > ali.sstart and ali.qstop < ali.sstop:
										#TODO
										ali_str = str(ali).split()[:2]
										sys.stderr.write('...excluding alignment\n\t{}\n\t{}\n'.format(ali_str[0], ali_str[1]))
										ali_count-=1
									elif ali.sstart > ali.qstart and ali.sstart< ali.qstop:
										#TODO
										ali_str = str(ali).split()[:2]
										sys.stderr.write('...excluding alignment\n\t{}\n\t{}\n'.format(ali_str[0], ali_str[1]))
										ali_count-=1
									else: pangenome.add_alignment(ali)
								else: pangenome.add_alignment(ali)
				new_dict['q'] = {"start" : [] , "stop" : [], "seq" : [] }
				new_dict['s'] = {"start" : [] , "stop" : [], "seq" : [] }
			if sbjct and line[:5] == 'Query':
				new_dict['q']['start'].append(int(line_list[1]))
				new_dict["q"]["seq"].append(line_list[-2])
				new_dict["q"]["stop"].append(int(line_list[-1]))
			if line[:5] == 'Sbjct':
				new_dict["s"]["start"].append(int(line_list[1]))
				new_dict["s"]["seq"].append(line_list[-2])
				new_dict["s"]["stop"].append(int(line_list[-1]))
	if new_dict['q']['start'] and (self_alignments or query!=sbjct): 
		q_pos_tuple = (new_dict["q"]["start"][0],new_dict["q"]["stop"][-1] )
		s_pos_tuple = (new_dict["s"]["start"][0], new_dict["s"]["stop"][-1])
		if max(q_pos_tuple) - min(q_pos_tuple) +1 >= min_ali_length and max(s_pos_tuple) - min(s_pos_tuple) +1 >= min_ali_length:
			if query != sbjct or q_pos_tuple != s_pos_tuple:
				ali_ind+=1
				ali = Alignment(query, sbjct, q_pos_tuple, s_pos_tuple, (''.join(new_dict["q"]["seq"]), ''.join(new_dict["s"]["seq"])))
				pangenome.add_alignment(ali)
	sys.stderr.write('...found {} valid pairwise alignments\n'.format(ali_ind))
	return pangenome


if __name__ == "__main__":
	pass




import sys
from pancake.fi import FeatureInstance
from pancake.sf import SharedFeature
from pancake.chromosome import Chromosome
from pancake.editops import EditOps,editops_from_ali
from pancake.utils import rev_comp,needle_wunsch,turn_ops,ali_refinement,new_ops_from_ops,ops_from_ops_and_ali


pos_interval = 10000

class PanGenome:

	def __init__(self):
		self.features = set() #set of Shared Features, holds all labeled FI or part of an alignment
		self.unlabeled = {} # foreach chromosome: SET of unlabeled FIs  
		self.chromosome_stops = {} # foreach chromosome (key): stop FI
		self.chrom_names = {} # foreach chromosome_name (key): a Chromosome
		self.genomes = {} #foreach genome: SET of chromosomes
		self.chromosomes = {} #foreach chromosome: a dictionary of start positions

	def __str__(self):
		out = ''
		for chrom in self.chromosomes:
			out += str(chrom) + '\n' + len(str(chrom)) * '-' + '\n'
			unlabeled_list = list(self.unlabeled[chrom])
			unlabeled_list.sort(key=lambda x: x.start)
			for fi in unlabeled_list.sort:
				out += str(fi) + '\n'
			out += '#\n'
		out += '#-#-#\n\n'
		for sf in self.features:
			out+= str(sf) +'\n'
		return out

	def add_chromosome(self, name, seq, genome=None):
		if not genome: genome=name
		chrom = Chromosome(name, genome)
		if genome not in self.genomes: self.genomes[genome] = set()
		self.genomes[genome].add(chrom)
		self.chrom_names[name] = chrom
		fi = FeatureInstance( chrom, 1, len(seq), seq.upper(), None, None, None )
		self.chromosome_stops[chrom] = fi
		self.unlabeled[chrom] = set([fi])
		self.chromosomes[chrom] = {}
		for pos_ind in range(int(self.chrom_length(chrom)/pos_interval) +1):
			self.chromosomes[chrom][pos_ind * pos_interval +1] = fi

	def add_alignment(self, ali):
		while ali: 
			#sys.stderr.write(self.seq(self.chrom_names['gi|372116212|gb|CP003216.1|'], 2178033, 2178035) +'\n')
			ali = self.add_ali(ali)
		return self

	def add_ali(self, ali):
		qchrom = self.chrom_names[ali.qname]
		schrom = self.chrom_names[ali.sname]
		if ali.qstop-ali.qstart > -1 and ali.sstop-ali.sstart > -1:
			q_init_fi, q_overlap, s_init_fi, s_overlap = self.start_parameters(ali)
		
			#treat initial q_fi (at ali begin)
			if self.treat_init_fi_at_ali_begin(ali.qstrand, q_init_fi, q_overlap): 
				q_init_fi, q_overlap, s_init_fi, s_overlap = self.start_parameters(ali)
			#treat initial s_fi (at ali begin)
			if self.treat_init_fi_at_ali_begin(ali.sstrand, s_init_fi, s_overlap): 
				q_init_fi, q_overlap, s_init_fi, s_overlap = self.start_parameters(ali)

			#check alignmnets at alignment end
			if self.treat_init_fi_at_ali_end(ali.qstrand, q_init_fi, q_overlap): 
				q_init_fi, q_overlap, s_init_fi, s_overlap = self.start_parameters(ali)
			if self.treat_init_fi_at_ali_end(ali.sstrand, s_init_fi, s_overlap): 
				q_init_fi, q_overlap, s_init_fi, s_overlap = self.start_parameters(ali)

			qcut, scut = 0, 0
			if ali.qstrand and q_overlap[1] > 0: qcut = q_overlap[1]
			if not ali.qstrand and q_overlap[0] > 0: qcut = q_overlap[0]

			if ali.sstrand and s_overlap[1] > 0: scut =s_overlap[1]
			if not ali.sstrand and s_overlap[0] > 0: scut = s_overlap[0]
			#print('qcut', qcut, 'scut', scut)
			if qcut == 0:
				if scut == 0:
					if not q_init_fi.sh_feat or not s_init_fi.sh_feat or q_init_fi.sh_feat != s_init_fi.sh_feat:
						self.merge_sfs(ali, q_init_fi, s_init_fi)
				else:
					#scut>0: find cut_index in alignment
					#always from alignment start, marks #bases in first alignment part
					ali_stop_ind = s_init_fi.stop - s_init_fi.start+1
					ali1, ali2 = ali.divide(ali_stop_ind, False)

					#cut q_init_fi
					if ali1: # and ali1.sstop -ali1.sstart > -1 and ali1.qstop -ali1.qstart > -1:
						self.add_ali(ali1)

					if ali2:
						return ali2 #self.add_alignment(ali2)
					
			else:
				#qcut>0
				if scut == 0:
					ali_stop_ind = q_init_fi.stop - q_init_fi.start+1
					ali1, ali2 = ali.divide(ali_stop_ind, True)
					if ali1:
						self.add_ali(ali1)
					if ali2:
						return ali2 #self.add_alignment(ali2)

				else:
					#qcut>0 and scut>0
					#find cut_ind
					q_stop_ind = q_init_fi.stop - q_init_fi.start + 1
					s_stop_ind = s_init_fi.stop -s_init_fi.start + 1

					i, c1, c2 = 0, 0, 0
					while (c1 < q_stop_ind and c2 < s_stop_ind):
						c1 += ali.qali[i] != '-'
						c2 += ali.sali[i] != '-'
						i += 1
					cut1 = len(ali.qali[:i].replace('-', ''))
					cut2 = len(ali.sali[:i].replace('-', ''))

					if cut1 >0:
						ali1, ali2 = ali.divide(cut1, True)
					elif cut2>0:
						ali1, ali2 = ali.divide(cut2, False)
					else: ali1, ali2 = None, None 

					while ali1:
						ali1 = self.add_ali(ali1)

					if ali2:
						return ali2
		else: return None


	#cut initial fis at alignment begin, strand = ali.strand
	def treat_init_fi_at_ali_begin(self, strand, fi, overlaps):
		treated=False
		fi_strand = not fi.seq.reverse if fi.sh_feat else True
		if strand:
			if overlaps[0] <0:
				if fi_strand:
					self.treat_start(fi, abs(overlaps[0]))
				else:
					self.treat_stop(fi, abs(overlaps[0]))
				treated =True
		else:
			#start_fi is stop_fi!
			if overlaps[1] < 0:
				if fi_strand:
					self.treat_stop(fi, abs(overlaps[1]))
				else:
					self.treat_start(fi, abs(overlaps[1]))
				treated = True
		return treated

	#cut initial fis at alignment end
	def treat_init_fi_at_ali_end(self, strand, fi, overlaps):
		treated=False
		fi_strand = not fi.seq.reverse if fi.sh_feat else True
		if strand:
			if overlaps[1] <0:
				if fi_strand:
					self.treat_stop(fi, abs(overlaps[1]))
				else:
					self.treat_start(fi, abs(overlaps[1]))
				treated =True
		else:
			if overlaps[0] < 0:
				if fi_strand:
					self.treat_start(fi, abs(overlaps[0]))
				else:
					self.treat_stop(fi, abs(overlaps[0]))
				treated = True
		return treated


	# ACGTAG, overlap=2 --> AC|GTAG
	def treat_start(self, ref_fi, overlap):
		if ref_fi.sh_feat:
			fi1, fi2 = self.treat_labeled_overlap( ref_fi, overlap)
		else:
			fi1, fi2 = self.treat_unlabeled_overlap(ref_fi, overlap)
		return fi1, fi2


	def treat_labeled_overlap(self, ref_fi, overlap):

		#make sure that ref_fi is in + direction
		if ref_fi.seq.reverse: 
			ref_fi.sh_feat.turn()
			overlap = ref_fi.stop-ref_fi.start - overlap +1
		#cut_ind = ref_fi.seq.ref_cut_ind(overlap) # cutting index of reference sequence in shared feature index: should not be applied for ref_fi!
		cut_ind, ref_ops1,ref_ops2 = ref_fi.seq.ref_cut_ind_and_ref_ops(overlap)
		sf = ref_fi.sh_feat
		new_sf1 = SharedFeature(sf.seq[:cut_ind])
		new_sf2 = SharedFeature(sf.seq[cut_ind:])
		return_fi1, return_fi2 = None, None
		for fi in list(sf):
			#reference feature instance is forward
			new_edit_ops = fi.seq.divide(cut_ind) if fi!=ref_fi else (ref_ops1,ref_ops2)
			l1 = len(new_edit_ops[0])
			l2 = len(new_edit_ops[1])
			if l1 >0 and l2 >0:
				if not fi.seq.reverse:
					new_fi1 = FeatureInstance(fi.chrom, fi.start, fi.start+l1-1, new_edit_ops[0], new_sf1, None, fi.prevFI)
					new_sf1.add(new_fi1)

					new_fi2 = FeatureInstance(fi.chrom, fi.start+l1, fi.stop , new_edit_ops[1], new_sf2, fi.nextFI, new_fi1)
					new_sf2.add(new_fi2)
					
					new_fi1.nextFI = new_fi2
					
					if new_fi1.prevFI:
						new_fi1.prevFI.nextFI = new_fi1
					if new_fi2.nextFI:
						new_fi2.nextFI.prevFI = new_fi2
					else:
						self.chromosome_stops[new_fi2.chrom] = new_fi2
				#reverse fi
				else:
					new_fi1 = FeatureInstance(fi.chrom, fi.stop-l1+1, fi.stop, new_edit_ops[0], new_sf1, fi.nextFI, None)
					new_sf1.add(new_fi1)
					new_fi2 = FeatureInstance(fi.chrom, fi.start, fi.stop-l1, new_edit_ops[1],new_sf2, new_fi1, fi.prevFI )
					new_sf2.add(new_fi2)
				
					new_fi1.prevFI = new_fi2
					if fi.prevFI:
						fi.prevFI.nextFI = new_fi2
					if fi.nextFI:
						fi.nextFI.prevFI = new_fi1
					else:
						self.chromosome_stops[fi.chrom] = new_fi1
				if fi==ref_fi: return_fi1, return_fi2 = new_fi1, new_fi2
				#update self.chromosomes
				start, stop = new_fi1.start, new_fi1.stop
				iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
				iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
				for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
					self.chromosomes[new_fi1.chrom][pos] = new_fi1
				start, stop = new_fi2.start, new_fi2.stop
				iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
				iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
				for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
					self.chromosomes[new_fi2.chrom][pos] = new_fi2

			elif l1 > 0:
				#print('l1>0')
				fi.seq = new_edit_ops[0]
				fi.sh_feat = new_sf1
				new_sf1.add(fi)
				if fi==ref_fi: return_fi1 = fi
			elif l2 > 0:
				#print('l2>0')
				fi.seq = new_edit_ops[1]
				fi.sh_feat = new_sf2
				new_sf2.add(fi)
				if fi==ref_fi: return_fi2 = fi
			else:
				break
		if new_sf1 and len(new_sf1.seq) == 0:
			new_sf1.treat_empty_seq()
		if new_sf2 and len(new_sf2.seq) == 0:
			new_sf2.treat_empty_seq()
		self.features.add(new_sf1)
		self.features.add(new_sf2)
		try:
			self.features.remove(sf)
		except:
			sys.stderr.write('Tried to remove\n{}\n{}\n{}\n{}\n'.format(sf, ref_fi.sh_feat, return_fi1.sh_feat, return_fi2.sh_feat))
		#TODO check if necessary
		#self.check_non_similarity(new_sf1)
		#self.check_non_similarity(new_sf2)
		return return_fi1, return_fi2

	def treat_unlabeled_overlap(self, ref_fi, overlap):
			#unlabeled --> unlabeled (ref_fi, l=overlap) + unlabeled (new_fi)
			ref_chrom = ref_fi.chrom
			self.unlabeled[ref_chrom].remove(ref_fi)
			new_fi = FeatureInstance(ref_chrom, ref_fi.start+abs(overlap), ref_fi.stop, ref_fi.seq[overlap:], None, ref_fi.nextFI, ref_fi)
			if new_fi.nextFI:
				new_fi.nextFI.prevFI = new_fi
			else:
				self.chromosome_stops[ref_chrom] = new_fi
			ref_fi.stop = new_fi.start-1
			ref_fi.seq = ref_fi.seq[:overlap]
			ref_fi.nextFI = new_fi
			if ref_fi.prevFI:
				ref_fi.prevFI.nextFI = ref_fi

			self.unlabeled[ref_chrom].add(ref_fi)
			self.unlabeled[ref_chrom].add(new_fi)
			
			start, stop = ref_fi.start, ref_fi.stop
			iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
			iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
			for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
				self.chromosomes[ref_chrom][pos] = ref_fi
			start, stop = new_fi.start, new_fi.stop
			iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
			iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
			for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
				self.chromosomes[ref_chrom][pos] = new_fi
			return ref_fi, new_fi

	# ACGTAG, overlap=2 --> ACGT|AG
	def treat_stop(self, ref_fi, overlap):
		#treat stop is nothing else than treat_start with adapted parameters
		if ref_fi.sh_feat:
			fi1, fi2 = self.treat_labeled_overlap(ref_fi, ref_fi.stop-overlap-ref_fi.start+1)
		else:
			fi1, fi2 = self.treat_unlabeled_overlap(ref_fi, ref_fi.stop-overlap-ref_fi.start+1)
		return fi1, fi2

	def start_parameters(self, ali):
		#overlap[0] always concerns start parameters, overlap[1] always concerns stop parameters, regardless of strand
		#query
		if ali.qstrand:
			q_fi =  self.crossing_fi(self.chrom_names[ali.qname], ali.qstart)
		else:
			q_fi =  self.crossing_fi(self.chrom_names[ali.qname], ali.qstop)
		q_overlap = (q_fi.start-ali.qstart, ali.qstop - q_fi.stop)
		#subject
		if ali.sstrand:
			s_fi = self.crossing_fi(self.chrom_names[ali.sname], ali.sstart)
		else:
			s_fi = self.crossing_fi(self.chrom_names[ali.sname], ali.sstop)
		s_overlap = (s_fi.start-ali.sstart, ali.sstop - s_fi.stop)
		return (q_fi, q_overlap, s_fi, s_overlap)

	def crossing_fi(self, chrom, pos):
		index = (pos-1) // pos_interval * pos_interval +1
		if pos -index < index + pos_interval -pos or index+pos_interval not in self.chromosomes[chrom]:
			start_fi = self.chromosomes[chrom][index]
			while(start_fi.stop < pos) and start_fi.nextFI:
				start_fi = start_fi.nextFI
		else:
			start_fi = self.chromosomes[chrom][index+pos_interval]
			while(start_fi.start > pos) and start_fi.prevFI:
				start_fi = start_fi.prevFI
		return start_fi

	def chrom_length(self, chrom):
		return self.chromosome_stops[chrom].stop

	def seq(self, chrom, start=1, stop=None, start_fi =None):
		#catch ERROR indices and ERROR genome
		if not stop: stop = self.chrom_length(chrom)
		if stop > self.chrom_length(chrom): 
			sys.stderr.write('While Sequence Retrieval: Stop position {} not in {}. Set to chromosome length = {}.\n'.format(stop, chrom, self.chrom_length(chrom)))
		#get start_fi
		if not start_fi: 
			start_fi = self.crossing_fi(chrom, start)
		begin_olp = start_fi.start-start
		seq_list = []
		seq_list.append(start_fi.sequence()[abs(begin_olp):])
		end_olp = stop-start_fi.stop
		if end_olp<0:
			seq_list[0] = seq_list[0][:end_olp]
		elif end_olp>0:
			while start_fi.stop < stop:
				start_fi = start_fi.nextFI
				seq_list.append(start_fi.sequence())
				if start_fi.stop > stop:
					seq_list[-1] = seq_list[-1][:-(start_fi.stop-stop)]
		return ''.join(seq_list)


	def merge_sfs(self, ali, q_start_fi, s_start_fi):
		#FI/SF handling
		if not q_start_fi.sh_feat:
			q_start_fi = self.convert_to_sf(q_start_fi)
		if not s_start_fi.sh_feat:
			s_start_fi = self.convert_to_sf(s_start_fi)

		if len(s_start_fi.sh_feat) > len(q_start_fi.sh_feat):
			ali.convert()
			q_start_fi, s_start_fi = s_start_fi, q_start_fi

		qsf = q_start_fi.sh_feat #this will get more entries
		old_sf = s_start_fi.sh_feat # this will be deleted

		if ali.qstrand and q_start_fi.seq.reverse: qsf.turn()
		elif not ali.qstrand and not q_start_fi.seq.reverse: qsf.turn()
		q_ref_ali = q_start_fi.seq.get_alignment(qsf.seq) #alignment of qsf.seq and q_start_fi.seq

		#ali.sali and s_start_fi are in same direction
		if ali.sstrand and s_start_fi.seq.reverse: old_sf.turn()
		elif not ali.sstrand and not s_start_fi.seq.reverse: old_sf.turn()


		#qali is always in right direction
		qsf_seq_ops = ops_from_ops_and_ali(list(q_start_fi.seq.ops), qsf.seq, ali.qali, ali.sali) #ops of ali.sali rebased to qsf.seq
		qsf_seq_ops = ali_refinement(qsf.seq, qsf_seq_ops)

		sseq = ali.sali.replace('-','')
		turned_ops = turn_ops(qsf.seq, list(qsf_seq_ops)) # get editops for qsf.seq based on s_start_fi.seq
		if len(old_sf) >1:
			ali_dic = old_sf.alignments_to_entry(s_start_fi) #alignment of s_start_fi.seq and entry
			for entry in ali_dic:
				other_ops = new_ops_from_ops(list(turned_ops), ali_dic[entry], sseq)
				other_ops = ali_refinement(qsf.seq, other_ops)
				entry.sh_feat = qsf
				entry.seq.ops = other_ops
				qsf.add(entry)
				if entry.nextFI:
					entry.nextFI.prevFI = entry
				else: 
					self.chromosome_stops[entry.chrom] = entry
				if entry.prevFI:
					entry.prevFI.nextFI = entry
				#update pancake.chromosomes
				start, stop = entry.start, entry.stop
				iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
				iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
				for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
					self.chromosomes[entry.chrom][pos] = entry
		s_start_fi.sh_feat = qsf
		qsf.add(s_start_fi)
		s_start_fi.seq.ops = qsf_seq_ops
		if s_start_fi.nextFI: 
			s_start_fi.nextFI.prevFI = s_start_fi
		else:
			self.chromosome_stops[s_start_fi.chrom] = s_start_fi
		if s_start_fi.prevFI:
			s_start_fi.prevFI.nextFI = s_start_fi
		#index update
		start, stop = s_start_fi.start, s_start_fi.stop
		iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
		iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
		for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
			self.chromosomes[s_start_fi.chrom][pos] = s_start_fi
		self.features.remove(old_sf)


	def convert_to_sf(self, fi):
		#convert unlabeled fi into shared feature
		self.unlabeled[fi.chrom].remove(fi)
		new_sf = SharedFeature(fi.seq.upper(), {fi})
		fi.seq = EditOps([fi.stop-fi.start+1])
		
		fi.sh_feat = new_sf
		if fi.nextFI: 
			fi.nextFI.prevFI = fi
		else:
			self.chromosome_stops[fi.chrom] = fi
		if fi.prevFI:
			fi.prevFI.nextFI = fi
		#update pancake.chromosomes
		start, stop = fi.start, fi.stop
		iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
		iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
		for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
			self.chromosomes[fi.chrom][pos] = fi
		self.features.add(new_sf)
		return fi


###PanCake Refinement

	def refine_adjacent_sfs(self):
	
		#check unlabeled
		for chrom in self.unlabeled:
			concatenation = True
			while concatenation:
				entry_list = list(self.unlabeled[chrom])
				concatenation =False
				for entry in entry_list:
					if entry.nextFI in entry_list:
						self.unlabeled[chrom].remove(entry.nextFI)
						entry.seq += entry.nextFI.seq
						entry.stop = entry.nextFI.stop
						entry.nextFI = entry.nextFI.nextFI
						if entry.nextFI:
							entry.nextFI.prevFI = entry
						else:
							self.chromosome_stops[chrom] = entry
						if entry.prevFI: entry.prevFI.nextFI = entry
						#index update
						start, stop = entry.start, entry.stop
						iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
						iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
						for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
							self.chromosomes[chrom][pos] = entry
						concatenation = True
						break
	
		'''join CSOs with perfectly adjacent entries'''
		#check sfs
		sf_list = list(self.features)
		while sf_list:
			sf = sf_list[0]
			sec_sf = None
			entry_list = list(sf.entries)
			start_entry, entry_list = entry_list[0], entry_list[1:]
			ok_flag = True
			if start_entry.seq.reverse: start_entry.sh_feat.turn()

			#check forward (nextFI.sh_feat)
			ref_fi = start_entry.nextFI
			if ref_fi and ref_fi.sh_feat:
				if ref_fi.seq.reverse: ref_fi.sh_feat.turn()
				next_sf = ref_fi.sh_feat
			else:
				ok_flag = False
			if ok_flag and next_sf and len(sf) == len(next_sf):
				for fi in entry_list:
					if not ok_flag: break
					if not fi.seq.reverse:
						next_fi = fi.nextFI
						if not next_fi or not next_fi.sh_feat: ok_flag = False
						elif next_fi.sh_feat and (next_fi.sh_feat != next_sf or next_fi.seq.reverse != fi.seq.reverse): ok_flag = False
					else:
						prev_fi = fi.prevFI
						if not prev_fi or not prev_fi.sh_feat: ok_flag=False
						elif prev_fi.sh_feat and (prev_fi.sh_feat != next_sf or prev_fi.seq.reverse != fi.seq.reverse): ok_flag = False
			else: ok_flag = False

			#if ok_flag: concatenate, update list
			if ok_flag:
				self.concatenate_sfs(sf, next_sf)
				if next_sf in sf_list: sf_list.remove(next_sf)
			else:
				#check backward
				ok_flag=True
				ref_fi = start_entry.prevFI
				if ref_fi and ref_fi.sh_feat:
					if ref_fi.seq.reverse : ref_fi.sh_feat.turn()
					prev_sf = ref_fi.sh_feat
				else: ok_flag = False
				
				if ok_flag and prev_sf and len(sf) == len(prev_sf):
					for fi in entry_list:
						if not fi.seq.reverse:
							prev_fi = fi.prevFI
							if not prev_fi or not prev_fi.sh_feat: ok_flag = False
							elif prev_fi.sh_feat and (prev_fi.sh_feat != prev_sf or prev_fi.seq.reverse != fi.seq.reverse): ok_flag = False
						else:
							next_fi = fi.nextFI
							if not next_fi or not next_fi.sh_feat: ok_flag=False
							elif next_fi.sh_feat and (next_fi.sh_feat != prev_sf or next_fi.seq.reverse != fi.seq.reverse): ok_flag = False
				else: ok_flag=False

				if ok_flag:
					self.concatenate_sfs(prev_sf, sf)
					if sf in sf_list: sf_list.remove(sf)
				else:
					#not adjacent, remove from list
					sf_list.remove(sf)


	def concatenate_sfs(self, sf, sec_sf):
		#check SFs
		entry_list = list(sf.entries)
		sf.seq = sf.seq + sec_sf.seq
		for fi in entry_list:
			if not fi.seq.reverse:
				fi2 = fi.nextFI
				fol_sf = fi2.sh_feat if fi2 else None
			else:
				fi2 = fi.prevFI
				fol_sf = fi2.sh_feat if fi2 else None
			if fol_sf != sec_sf:
				raise ValueError('This went wrong')
			if type(fi.seq.ops[-1]) == int and type(fi2.seq.ops[0])==int:
				if (fi.seq.ops[-1] > 0 and fi2.seq.ops[0] > 0) or (fi.seq.ops[-1] < 0 and fi2.seq.ops[0] < 0):
					new_edit_ops = EditOps(fi.seq.ops[:-1] + [fi.seq.ops[-1] + fi2.seq.ops[0]] + fi2.seq.ops[1:])
				else:
					new_edit_ops = EditOps(fi.seq.ops + fi2.seq.ops)
			else:
				new_edit_ops = EditOps(fi.seq.ops + fi2.seq.ops)
			if fi2.seq.reverse: new_edit_ops.reverse = True

			fi.seq = new_edit_ops
			if not fi.seq.reverse:
				next_fi = fi2.nextFI
				prev_fi = fi.prevFI
				fi.stop = fi2.stop
			else:
				next_fi = fi.nextFI
				prev_fi = fi2.prevFI
				fi.start = fi2.start
			fi.nextFI = next_fi
			fi.prevFI = prev_fi

			if prev_fi: 
				prev_fi.nextFI = fi
			if next_fi:
				next_fi.prevFI = fi
			else:
				self.chromosome_stops[fi.chrom] = fi
			
			start, stop = fi.start, fi.stop
			iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
			iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
			for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
				self.chromosomes[fi.chrom][pos] = fi

		self.features.remove(sec_sf)
		for fi in sec_sf:
			del (fi)
		del (sec_sf)


	def refine_unlabeled(self):
		#identifies Shared Features with only 1 entry and integrate them in self.unlabeled
		lonley_sfs=[]
		for sf in self.features:
			if len(sf) ==1: lonley_sfs.append(sf)
		for sf in lonley_sfs:
			ind, ufi_ind = -1, 0
			fi = list(sf.entries)[0]
			if fi.seq.reverse: fi.sh_feat.turn() #unlabeled fis are always 'forward'
			fi.seq = fi.seq.seq(fi.sh_feat.seq)
			self.features.remove(fi.sh_feat)
			fi.sh_feat = None
			if fi.nextFI:
				fi.nextFI.prevFI = fi
			else: self.chromosome_stops[fi.chrom] = fi
			if fi.prevFI:
				fi.prevFI.nextFI = fi
			if fi.prevFI in self.unlabeled[fi.chrom]:
				#concatenate fi with previous fi
				self.unlabeled[fi.chrom].remove(fi.prevFI)
				fi.start = fi.prevFI.start
				fi.seq = fi.prevFI.seq + fi.seq
				fi.prevFI = fi.prevFI.prevFI
				if fi.prevFI:
					fi.prevFI.nextFI = fi
				if fi.nextFI:
					fi.nextFI.prevFI = fi
				else: self.chromosome_stops[fi.chrom] = fi

			if fi.nextFI in self.unlabeled[fi.chrom]:
				#concatenate fi with next fi
				self.unlabeled[fi.chrom].remove(fi.nextFI)
				fi.stop = fi.nextFI.stop
				fi.seq += fi.nextFI.seq
				fi.nextFI = fi.nextFI.nextFI
				if fi.prevFI:
					fi.prevFI.nextFI = fi
				if fi.nextFI:
					fi.nextFI.prevFI = fi
				else: self.chromosome_stops[fi.chrom] = fi

			#update self.chromosomes
			start, stop = fi.start, fi.stop
			iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
			iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
			for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
				self.chromosomes[fi.chrom][pos] = fi
			
			self.unlabeled[fi.chrom].add(fi)

	def check_non_similarity(self):
		sfs=list(self.features)
		for sf in sfs:
			self.check_sf_for_non_similarity(sf)

	def check_sf_for_non_similarity(self, sf):
		tmp_fis = set([0])
		while tmp_fis:
			tmp_fis = set()
			for fi in sf.entries:
				similarity_flag = False
				for op in fi.seq.ops:
					if type(op) == int and op>0:
						similarity_flag = True
						break
				if not similarity_flag:
					#only substitutions, deletions and insertions in fi's editops
					tmp_fis.add(fi)

			if tmp_fis:
				new_sf = SharedFeature('')
				for fi in tmp_fis:
					fi.seq.ops = list(fi.seq.seq(sf.seq).upper())
					fi.sh_feat = new_sf
					sf.entries.remove(fi)
					new_sf.add(fi)
					if fi.nextFI:
						fi.nextFI.prevFI = fi
					else: self.chromosome_stops[fi.chrom] = fi
					if fi.prevFI:
						fi.prevFI.nextFI = fi
					#update self.chromosomes
					start, stop = fi.start, fi.stop
					iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
					iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
					for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
						self.chromosomes[fi.chrom][pos] = fi
				new_sf.treat_empty_seq()
				if len(sf.entries) == 0:
					self.features.remove(sf)
				self.features.add(new_sf)
				sf = new_sf


######PanCake Serialization

	def compress(self, pan_file):
		'''compress into given pan_file'''
		with open(pan_file, 'wt') as out_file:
			sf_dict = {}
			i = 1
			for sf in self.features:
				sf_dict[sf] = i
				out_file.write(sf.seq + '\n')
				i+=1
			for chrom in self.chromosomes:
				out_file.write('#' + chrom.name)
				for chr_name in self.chrom_names:
					if self.chrom_names[chr_name]==chrom and chr_name != chrom.name: out_file.write(',' + chr_name)

				out_file.write(',({})'.format(chrom.genome))
				out_file.write('\n')
				fi = self.chromosomes[chrom][1]
				while fi:
					if fi.sh_feat: fi.seq.treat_double_matches() #TODO avoid this
					out_file.write(str(fi.stop) + ' ' + str(fi.seq))
					if fi.sh_feat: out_file.write(' ' + str(sf_dict[fi.sh_feat]))
					out_file.write('\n')
					fi = fi.nextFI


def decompress(pan_file):
	fi_part = False
	sf_dict={}
	sf_count = 1
	prev_fi = None
	chrom = None

	pancake = PanGenome()
	with open(pan_file, 'rt') as in_file:
		for line in in_file:
			if line[0] == '#': fi_part=True #all SFs parsed
			if not fi_part:
				sf_dict[sf_count] = SharedFeature(line[:-1])
				pancake.features.add(sf_dict[sf_count])
				sf_count+=1
			else:
				if line[0] == '#':
					chrom_names = line[1:-1].split(',')
					genome = chrom_names[-1][1:-1]
					if genome not in pancake.genomes: pancake.genomes[genome] = set()
					chrom = Chromosome(chrom_names[0], genome)
					pancake.genomes[genome].add(chrom)
					pancake.chromosome_stops[chrom] = None
					pancake.chromosomes[chrom] = {}
					pancake.unlabeled[chrom] = set()
					prev_fi = None
					for chrom_name in chrom_names[:-1]: pancake.chrom_names[chrom_name] = chrom
				else:
					line_list = line[:-1].split(' ')
					if len(line_list) == 2:
						seq = line_list[1]
						sf=None
					else:
						if line_list[1][0] == '-':
							seq = EditOps(line_list[1][1:], True)
						else:
							seq = EditOps(line_list[1], False)
						sf=sf_dict[int(line_list[2])]
						seq.list_ops()
					
					if not prev_fi:
						fi = FeatureInstance(chrom, start=1, stop=int(line_list[0]), seq = seq, sh_feat=sf, nextFI=None, prevFI=None)
						pancake.chromosomes[chrom][1] = fi
					else:
						fi = FeatureInstance(chrom, start=prev_fi.stop+1, stop=int(line_list[0]), seq = seq, sh_feat=sf, nextFI=None, prevFI=prev_fi)
						prev_fi.nextFI = fi
					pancake.chromosome_stops[chrom] = fi
					prev_fi = fi
					if fi.sh_feat != None: sf.add(fi)
					else: pancake.unlabeled[chrom].add(fi)
	#update pancake.chromosomes
	for chrom in pancake.chromosomes:
		start_fi = pancake.chromosomes[chrom][1]
		while start_fi:
			start, stop = start_fi.start, start_fi.stop
			iter_start = (start//pos_interval)+1 if start>(start//pos_interval) * pos_interval+1 else start//pos_interval
			iter_stop = stop//pos_interval+1 if stop >= (stop//pos_interval) * pos_interval+1 else stop//pos_interval
			for pos in [i*pos_interval +1 for i in range(iter_start, iter_stop)]:
				pancake.chromosomes[chrom][pos] = start_fi
			start_fi = start_fi.nextFI
	return pancake


if __name__ == "__main__":
	pass



class Chromosome:

	def __init__(self, name, genome):
		self.name = name
		self.genome = genome #string 'GENOME_NAME'

	def __str__(self):
		'''will be used for pangenome output'''
		return self.name #+' (' + self.genome + ')'

	def new_name(self, new_name):
		'''rename Chromosome'''
		self.name = new_name

	def add_to_genome(self, genome):
		self.genome = genome





# coding: utf-8
from setuptools import setup
import sys

if sys.version_info < (3,2):
	sys.stdout.write("At least Python 3.2 is required.\n")
	sys.exit(1)

__version__ = '1.1.1'

setup(
	name = 'pancake',
	version = __version__,
	author = 'Corinna Ernst',
	author_email = 'corinna.ernst@uni-due.de',
	description = 'A Data Structure for Pangenomes -- Identification of Singletons and Core Regions Dependent on Pairwise Sequence Similarities',
	#long_description = "Non-databse approach for identification of singletons and core regions in arbitrary sequence sets dependent on pairwise sequence similarities (as provided by common sequence alignment tools), and independent of annotation data.'),
	license = 'MIT',
	url = 'https://bitbucket.org/CorinnaErnst/pancake',
	packages = ['pancake'],
	scripts = ['bin/pancake'],
	install_requires=["numpy", "biopython"],
	test_suite = 'nose.collector',
	classifiers = [
		"Development Status :: 4 - Beta",
		"Environment :: Console",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: MIT License",
		"Natural Language :: English",
		"Programming Language :: Python :: 3",
		"Topic :: Scientific/Engineering :: Bio-Informatics"
	]
)



#from pancake.utils import *
from pancake.utils import needle_wunsch

import sys

class EditOps:

	def __init__(self, ops, reverse=False):
		self.ops = ops # list of edit operations
		self.reverse = reverse #indicates if editoperation should be applied to ref seqquence or its reverse complement


	def __str__(self):
		if self.reverse:
			return '-' + ''.join(map(lambda x: -x*'.' if type(x)==int and x< 0 else str(x), self.ops))
		else:
			return ''.join(map(lambda x: -x*'.' if type(x)==int and x< 0 else str(x), self.ops))


	def __len__(self):
		'''return length of resulting sequence'''
		l = 0
		for op in self.ops:
			if type(op) == int:
				if op>0: l += op
			else: l += 1
		return l


	#used for tests and divide
	def ref_len(self):
		'''return length of corresponding reference sequence'''
		l = 0
		for op in self.ops:
			if type(op) == int:
				l+=abs(op)
			elif op.isupper():
				l += 1
		return l


	def list_ops(self):
		'''convert edit ops string into list'''
		l, dels = [], 0
		for i in self.ops:
			if dels and i != '.':
				l.append( -dels)
				dels=0
			if i.isdigit():
				if l and type(l[-1]) == str and l[-1].isdigit():
					l[-1] += i
				else:
					l.append(i)
			elif i == '.':
				dels += 1
			else:
				l.append(i)
		if dels:
			l.append( -dels)
		#convert matches into int types
		self.ops = list(map(lambda o: int(o) if type(o)==str and o.isdigit() else o, l))


	#used for turnSF
	def turn(self):
		self.ops = [x if type(x)==int else rev_comp(x) for x in self.ops[::-1]]
		self.reverse = not self.reverse


	def divide(self, cut_ind):
		'''return two new EditOps Objects, cut_ind represents cut_ind in corresponding ref_seq!'''
		'''call for NON_REFERENCE FeatureInstance in divideSF'''
		length = 0
		ops1, ops2 = [], [] #edit operation lists of new EditOps
		self_ops = self.ops
		for i in range(len(self_ops)):
			if length >= cut_ind:
				ops2 = self_ops[i:]
				break
			ops1.append(self_ops[i])
			#compute new length
			if type(self_ops[i]) == int:
				length += abs(self_ops[i])
			elif self_ops[i].isupper():
				length += 1
		op1, op2 = EditOps(ops1, self.reverse), EditOps(ops2, self.reverse)
		op1_ref_len = op1.ref_len()
		if op1_ref_len > cut_ind:
			if ops1[-1] > 0:
				op2.ops = [op1_ref_len-cut_ind] + ops2
				op1.ops[-1] = ops1[-1] - op1_ref_len + cut_ind #i.e. op1.ops[-1] = ops1[-1] - (op1_ref_len - cut_ind)
			else:
				#ops1 ends with deletion
				op2.ops = [cut_ind-op1_ref_len] + op2.ops #i.e. op2.ops = [-1*(op1_ref_len-cut_ind)] + op2.ops
				op1.ops[-1] = ops1[-1] + op1_ref_len - cut_ind # i.e. -1 * (ops1[-1] - (op1_ref_len - cut_ind))
		return op1, op2


	def ref_cut_ind_and_ref_ops(self, overlap):
		'''return cut_ind of reference sequence, overlap is always from the start of self.ops'''
		'''return cut_ind parameter for self.divide()'''
		c, lops,i = 0, 0,0
		for op in self.ops:
			if lops >= overlap:
				break
			i+=1
			if type(op) == int:
				if op > 0:
					c+=op
					lops+=op
				else:
					c+=abs(op)
			elif op.isupper():
				c+=1
				lops+=1
			elif op.islower():
				lops += 1
		new1,new2 = EditOps(self.ops[:i] ,self.reverse), EditOps(self.ops[i:], self.reverse)
		if lops>overlap:
			c = c-(lops-overlap)
			new1.ops[-1] = new1.ops[-1]-(lops-overlap)
			new2.ops = [lops-overlap] + new2.ops
		return c, new1,new2



	def seq(self, ref_seq):
		''' return dna sequence encrypted by edit_op '''
		seq=[]
		seq_ind = 0
		for item in self.ops:
			if type(item) == int:
				if item >0:
					#matches
					seq.append(ref_seq[seq_ind:seq_ind+item])
					seq_ind+=item
				else:
					#deletions
					seq_ind += abs(item)
			elif item.islower():
				#insertion
				seq.append(item.upper())
			else:
				#substitution
				seq.append(item)
				seq_ind += 1
		seq = ''.join(seq)
		return seq


	def get_alignment(self, ref_seq):
		'''return alignment'''
		ref_ali_seq = ''
		ali_seq = ''
		seq_ind = 0
		for item in self.ops:
			if type(item) == int:
				if item >0:
					#matches
					ref_ali_seq += ref_seq[seq_ind:seq_ind+item]
					ali_seq += ref_seq[seq_ind:seq_ind+item]
					seq_ind += item
				else:
					#deletions
					ref_ali_seq += ref_seq[seq_ind:seq_ind+abs(item)]
					ali_seq += abs(item) * '-'
					seq_ind += -item
			elif item.islower():
				#insertion
				ref_ali_seq += '-'
				ali_seq += item.upper()
			else:
				#substitution
				ref_ali_seq += ref_seq[seq_ind]
				ali_seq += item.upper()
				seq_ind += 1
		return (ref_ali_seq, ali_seq)


	#get rid of subsequent match ops
	#TODO avoid this call
	def treat_double_matches(self):
		i =1
		while i<len(self.ops):
			if (type(self.ops[i-1]), type(self.ops[i])) == (int,int) and self.ops[i-1]>0 and self.ops[i]>0:
				self.ops[i-1] += self.ops[i]
				self.ops = self.ops[:i] + self.ops[i+1:]
			else: i+=1


def editops_from_ali(q_string, s_string, reverse=True):
	'''return EditOps Instance computed from given alignment, EditOps represents sequence in s_string'''
	'''s_string is assumed to be already reverse complement, if reverse_flag set'''
	matches, deletions, ops = 0, 0, []
	for qs, ss in zip(q_string, s_string):
		if qs.upper() == ss.upper():
			matches+=1
			if deletions>0: 
				ops.append(-deletions)
				deletions=0
		else:
			if matches>0:
				ops.append(matches)
				matches=0
			if ss == '-':
				deletions+=1
			else:
				if deletions>0:
					ops.append(-deletions)
					deletions=0
				if qs == '-':
					ops.append(ss.lower())
				else:
					ops.append(ss.upper())
	if matches>0:
		ops.append(matches)
	if deletions >0: ops.append(-deletions)
	return EditOps( ops, reverse)


def editops_from_seqs(ref_seq, seq, reverse=False):
	'''return EditOps Instance computed from the alignment of the given sequences, EditOps represents seq'''
	'''if reverse: seq already have to be reverse complement'''
	q_string, s_string = needle_wunsch(ref_seq,seq)
	return editops_from_ali(q_string, s_string, reverse)

def rev_comp(s):
	return s.translate(str.maketrans('ACGTacgtRYMKrymkVBHDvbhd', 'TGCAtgcaYRKMyrkmBVDHbvdh'))[::-1]



if __name__ == "__main__":
	pass


